import os
from pathlib import Path

from qualibrate.app.api.exceptions.classes.storage import (
    QRelativeNotSubpathException,
)


def resolve_and_check_relative(base_path: Path, subpath: os.PathLike[str]) -> Path:
    """
    Build full path from base path and subpath. Raise error if build path isn't
    subpath of base path.

    Raises:
        ValueError: Built path isn't subpath of base path.
    """
    full = (base_path / Path(subpath)).resolve()
    if not full.is_relative_to(base_path):
        raise QRelativeNotSubpathException("Subpath isn't relative to base.")
    return full



from typing import Any, ClassVar


class Singleton(type):
    _instances: ClassVar[dict[type, type]] = {}

    def __call__(cls, *args: Any, **kwargs: Any) -> type:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]

    @classmethod
    def _clear(cls) -> None:
        cls._instances = {}



from collections.abc import Mapping
from pathlib import Path
from typing import Any

from pydantic import BaseModel


class Storage(BaseModel):
    path: Path | None = None
    data: Mapping[str, Any] | None = None



from collections.abc import Mapping, MutableMapping
from typing import Any

from jsonpointer import resolve_pointer


def recursive_properties_solver(
    properties: MutableMapping[str, Any], schema: Mapping[str, Any]
) -> MutableMapping[str, Any]:
    """
    Recursively resolves and simplifies properties in a schema by handling
    `$ref` and `allOf` constructs.

    Args:
        properties: A mapping of property names to their structures, which
            may include references (`$ref`) or combined definitions (`allOf`).
        schema: The full schema mapping used to resolve references.

    Returns:
        The updated properties mapping with resolved references and simplified
        structures.
    """
    for name, structure in properties.items():
        if "allOf" in structure and len(structure["allOf"]) == 1 and isinstance(structure["allOf"][0], Mapping):
            structure.update(structure.pop("allOf")[0])
        if "$ref" in structure:
            resolved = resolve_pointer(schema, structure["$ref"][1:])
            # for objects, object json schema has propertieskey
            # then they are handled
            if "properties" in resolved:
                properties[name] = recursive_properties_solver(resolved["properties"], schema)
            else:
                # for objects like enums, enums don't have properties key,
                # but they have ref so they also need to be handled
                properties[name] = resolved
                if "default" in structure:
                    properties[name]["default"] = structure["default"]
    return properties



import json
import logging
from collections.abc import Generator
from datetime import datetime
from functools import partial
from itertools import islice
from pathlib import Path
from typing import Any, TextIO

from qualibrate_config.models import QualibrateConfig

from qualibrate.core.utils.logger_m import logger
from qualibrate.core.utils.logger_utils.filters import filter_log_date

__all__ = [
    "parse_log_line",
    "parse_log_line_with_previous",
    "get_logs_from_qualibrate_files",
    "get_logs_from_qualibrate_in_memory_storage",
]

default_asctime_log_format = "%Y-%m-%d %H:%M:%S,%f"


def parse_log_line(line: str, previous_msg: dict[str, Any] | None = None) -> dict[str, Any]:
    # JSON format
    if line.startswith("{"):
        try:
            data = dict(json.loads(line))
        except json.decoder.JSONDecodeError as ex:
            logging.exception(f"Can't parse line {line}", exc_info=ex)
            return {}
        if "asctime" in data:
            data["asctime"] = datetime.strptime(data["asctime"], default_asctime_log_format)
        return data
    # old default-string-format
    parts = line.split(" - ", maxsplit=3)
    if len(parts) == 4:
        return {
            "asctime": datetime.strptime(parts[0], default_asctime_log_format),
            "name": parts[1],
            "levelname": parts[2],
            "message": parts[3].rstrip(),
        }
    if previous_msg is None:
        return {"message": line.rstrip()}
    copied = previous_msg.copy()
    copied["message"] = line.rstrip()
    return copied


def parse_log_line_with_previous(
    file: TextIO,
) -> Generator[dict[str, Any], None, None]:
    previous = None
    for line in file:
        previous = parse_log_line(line, previous)
        try:
            yield previous
        except TypeError as e:
            logging.exception(line, exc_info=e)


def get_logs_from_qualibrate_files(
    after: datetime | None = None,
    before: datetime | None = None,
    num_entries: int = 100,
    *,
    config: QualibrateConfig,
) -> list[dict[str, Any]]:
    log_folder = config.log_folder
    if log_folder is None:
        return []
    out_logs: list[dict[str, Any]] = []
    q_log_files = filter(Path.is_file, log_folder.glob("qualibrate.log*"))
    filter_log_date_range = partial(filter_log_date, after=after, before=before)
    for log_file in sorted(q_log_files):
        with open(log_file) as f:
            filtered = list(filter(filter_log_date_range, parse_log_line_with_previous(f)))
            lines_date_filtered = reversed(filtered)
            file_logs = islice(lines_date_filtered, num_entries - len(out_logs))
            out_logs.extend(file_logs)
            if len(out_logs) == num_entries:
                return list(reversed(out_logs))
    return list(reversed(out_logs))


def get_logs_from_qualibrate_in_memory_storage(
    after: datetime | None = None,
    before: datetime | None = None,
    num_entries: int = 100,
    *,
    config: QualibrateConfig,
) -> list[dict[str, Any]]:
    return logger.in_memory_handler.get_logs(after, before, num_entries)



import logging
from typing import cast


class EndpointFilter(logging.Filter):
    """Filter class to exclude specific endpoints from log entries."""

    def __init__(
        self,
        name: str = "",
        *,
        excluded_endpoints_starts: tuple[str, ...],
        success_status_codes: tuple[int, ...] | None = None,
    ) -> None:
        """
        Initialize the EndpointFilter class.

        Args:
            excluded_endpoints: A list of endpoints to be excluded from log
                entries.
        """
        super().__init__(name)
        self.excluded_endpoint_starts = excluded_endpoints_starts
        self.success_status_codes = success_status_codes or (200, 204)

    def filter(self, record: logging.LogRecord) -> bool:
        """
        Filter out log entries for excluded endpoints.

        Args:
            record: The log record to be filtered.

        Returns:
            bool: True if the log entry should be included, False otherwise.
        """
        if not record.args or len(record.args) < 5:
            return True

        # Get the endpoint path and status code from the log record
        # (expected to be the 3rd argument)
        log_args = cast(tuple[object, ...], record.args)
        endpoint_path = cast(str, log_args[2])
        response_status_code = cast(int, log_args[4])
        # Check if the endpoint path starts with any of the excluded patterns
        for excluded_pattern in self.excluded_endpoint_starts:
            if not endpoint_path.startswith(excluded_pattern):
                continue
            if response_status_code in self.success_status_codes:
                return False

        return True



import copy
import inspect
import sys
import traceback
from collections.abc import Generator, Mapping, Sequence
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
)

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends import (  # type: ignore[attr-defined]
    BackendFilter,
    backend_registry,
)
from pydantic import ValidationError, create_model
from qualibrate_config.resolvers import (
    get_qualibrate_config,
    get_qualibrate_config_path,
)

from qualibrate.core.config.resolvers import get_quam_state_path
from qualibrate.core.models.outcome import Outcome
from qualibrate.core.models.run_mode import RunModes
from qualibrate.core.models.run_summary.base import BaseRunSummary
from qualibrate.core.models.run_summary.node import NodeRunSummary
from qualibrate.core.models.run_summary.run_error import RunError
from qualibrate.core.parameters import NodeParameters
from qualibrate.core.q_runnnable import (
    QRunnable,
    file_is_calibration_node_instance,
    run_modes_ctx,
)
from qualibrate.core.runnables.run_action.action import ActionCallableType
from qualibrate.core.runnables.run_action.action_manager import (
    ActionDecoratorType,
    ActionManager,
)
from qualibrate.core.runnables.runnable_collection import RunnableCollection
from qualibrate.core.storage import StorageManager
from qualibrate.core.storage.local_storage_manager import LocalStorageManager
from qualibrate.core.utils.exceptions import StopInspection
from qualibrate.core.utils.logger_m import (
    ALLOWED_LOG_LEVEL_NAMES,
    LOG_LEVEL_NAMES_TYPE,
    logger,
)
from qualibrate.core.utils.node.comined_method import InstanceOrClassMethod
from qualibrate.core.utils.node.content import (
    parse_node_content,
    read_node_content,
    read_node_data,
)
from qualibrate.core.utils.node.loaders.base_loader import BaseLoader
from qualibrate.core.utils.node.path_solver import (
    get_node_dir_path,
)
from qualibrate.core.utils.node.record_state_update import update_node_machine
from qualibrate.core.utils.read_files import get_module_name, import_from_path
from qualibrate.core.utils.type_protocols import MachineProtocol, TargetType

__all__ = [
    "QualibrationNode",
    "NodeCreateParametersType",
    "NodeRunParametersType",
    "NodeMachineType",
]

NodeCreateParametersType = NodeParameters
NodeRunParametersType = NodeParameters
ParametersType = TypeVar("ParametersType", bound=NodeParameters)
NodeMachineType = TypeVar("NodeMachineType", bound=MachineProtocol)


def simplify_traceback(
    tb: Any,  # types.TracebackType
    node_filepath: Path,
) -> list[str]:
    """
    Simplify traceback to show only frames from the user's node code onwards.

    This removes framework overhead while preserving:
    - User's node code (action function entry point)
    - Subroutines called from the node (nested calls)
    - Libraries called from the node
    - Sub-libraries called by those libraries

    The algorithm finds the entry point into user code by looking for the
    first node file frame AFTER any action framework code. This handles:
    - Action errors with subroutines: Includes entire call chain within node
    - Action errors without subroutines: Starts from action function
    - Body errors: Starts from node body code
    - Import errors: Starts from import-time code

    Args:
        tb: The exception's traceback
        node_filepath: Path to the node file being executed

    Returns:
        List of formatted traceback strings
    """
    # Get all traceback frames
    all_frames = traceback.extract_tb(tb)

    # Find ALL frames in the node file
    node_filepath_str = str(node_filepath)
    node_frame_indices = [idx for idx, frame in enumerate(all_frames) if frame.filename == node_filepath_str]

    if not node_frame_indices:
        # Node file not in traceback (error during framework code
        # before node execution). Keep the full traceback
        return traceback.format_list(all_frames)

    # Find ALL frames in action framework code
    action_framework_indices = [
        idx
        for idx, frame in enumerate(all_frames)
        if "action_manager.py" in frame.filename or "action.py" in frame.filename
    ]

    if not action_framework_indices:
        # No action framework code - this is a node body error
        # (not inside an action) or import error
        # Start from the first node file occurrence
        start_idx = node_frame_indices[0]
    else:
        # Find the first node file frame AFTER the last action framework frame
        # This is the entry point into user code (action function)
        last_framework_idx = max(action_framework_indices)

        node_frames_after_framework = [idx for idx in node_frame_indices if idx > last_framework_idx]

        # Start from the first node frame after framework code
        # This captures the action function + all nested calls
        # Fallback: use first node frame
        # (unlikely edge case - shouldn't happen in normal execution)
        start_idx = node_frames_after_framework[0] if node_frames_after_framework else node_frame_indices[0]

    # Extract relevant frames from start point onwards
    relevant_frames = all_frames[start_idx:]

    # Format the frames
    return traceback.format_list(relevant_frames)


class QualibrationNode(
    QRunnable[ParametersType, ParametersType],
    Generic[ParametersType, NodeMachineType],
):
    """
    Represents a qualibration node (that can be run independently or as part
    of graph), responsible for executing specific tasks with defined parameters
     and modes.

    Args:
        name: The name of the node. If not provided, the name will be the name
            of the file that contains the node.
        parameters: Parameters passed to the node for its initialization.
            Defaults to None.
        description: A description of the node. Defaults to None.
        parameters_class: Class used for node parameters validation. Defaults
            to None.
        modes: Execution modes. Defaults to None.

    Raises:
        StopInspection: Raised if the node is instantiated in inspection mode.
    """

    active_node: Optional["QualibrationNode[ParametersType, Any]"] = None

    def __new__(cls, *args: Any, **kwargs: Any) -> "QualibrationNode[ParametersType, Any]":
        if cls.active_node is not None:
            return cls.active_node
        return super().__new__(cls)

    def __init__(
        self,
        name: str | None = None,
        parameters: ParametersType | None = None,
        description: str | None = None,
        *,
        parameters_class: type[ParametersType] | None = None,
        modes: RunModes | None = None,
        machine: NodeMachineType | None = None,
    ):
        if self.__class__.active_node is not None:
            return
        name = name or self.__class__._get_name_from_stack_frame()
        logger.info(f"Creating node {name}")
        parameters = self.__class__._validate_passed_parameters_options(name, parameters, parameters_class)
        super().__init__(
            name,
            parameters,
            description=description,
            modes=modes,
        )
        # class is used just for passing reference to the running instance
        self._fraction_complete = 0.0
        self.results: dict[Any, Any] = {}
        self.storage_manager: StorageManager[Self] | None = None
        self.machine = machine

        # Initialize the ActionManager to handle run_action logic.
        self._action_manager = ActionManager()
        self.namespace: dict[str, Any] = {}
        if self.modes.inspection:
            raise StopInspection("Node instantiated in inspection mode", instance=self)
        self._post_init()

    @staticmethod
    def _get_name_from_stack_frame() -> str:
        stack = inspect.stack()
        if not len(stack) or len(stack) < 3:
            raise ValueError("Can't resolve node name from node filename")
        frame = stack[2]
        return Path(frame.filename).stem

    def _post_init(self) -> None:
        self.run_start = datetime.now().astimezone()
        self.last_saved_at: datetime | None = None
        self._custom_action_label: str | None = None
        self._get_storage_manager()

        self._warn_if_external_and_interactive_mpl()

    @classmethod
    def _validate_passed_parameters_options(
        cls,
        name: str,
        parameters: ParametersType | None,
        parameters_class: type[ParametersType] | None,
    ) -> ParametersType:
        """
        Validates passed parameters and parameters class.

        If parameters
        passed then the instance will be used. If parameters class is passed,
        an attempt will be made to instantiate it. If neither parameters nor
        parameter class are passed, then the default base parameters will be
        used.

        Args:
            name: The name of the node.
            parameters: Parameters for the node.
            parameters_class: Parameters class.

        Returns:
            Validated parameters.

        Raises:
            ValueError: If parameters class instantiation fails.
        """
        params_type_error = ValueError("Node parameters must be of type NodeParameters")
        if parameters is not None:
            if not isinstance(parameters, NodeParameters):
                raise params_type_error
            if parameters_class is not None:
                logger.warning(
                    "Passed both parameters and parameters_class to the node "
                    f"'{name}'. Please use only parameters argument"
                )
            return parameters
        if parameters_class is None:
            fields = {name: copy.copy(field) for name, field in NodeParameters.model_fields.items()}
            # Create subclass of NodeParameters. It's needed because otherwise
            # there will be an issue with type checking of subclasses.
            # For example: NodeRunSummary.parameters
            new_model = create_model(  # type: ignore
                NodeParameters.__name__,
                __doc__=NodeParameters.__doc__,
                __base__=NodeParameters,
                __module__=NodeParameters.__module__,
                **{name: (info.annotation, info) for name, info in fields.items()},
            )
            return cast(ParametersType, new_model())
        logger.warning(
            f"parameters_class argument is deprecated. Please use parameters argument for initializing node '{name}'."
        )
        if not issubclass(parameters_class, NodeParameters):
            raise params_type_error
        try:
            return parameters_class()
        except ValidationError as e:
            raise ValueError(f"Can't instantiate parameters class of node '{name}'") from e

    @property
    def action_label(self) -> str | None:
        if self._custom_action_label is not None:
            return self._custom_action_label
        return self.current_action_name

    @action_label.setter
    def action_label(self, value: str | None) -> None:
        self._custom_action_label = value

    @property
    def machine(self) -> NodeMachineType | None:
        return self._machine

    @machine.setter
    def machine(self, value: NodeMachineType | None) -> None:
        self._machine = value
        self._machine_metadata: Mapping[str, Any] | None = None  # invalidate cache

    def __copy__(self) -> Self:
        """
        Creates a shallow copy of the node.

        This method copies the node, including its parameters, name, modes,
        and filepath, while resetting inspection-related modes to maintain
        consistency.

        Returns:
            A copy of the node.
        """
        modes = self.modes.model_copy(update={"inspection": False})
        active_node = self.__class__.active_node

        try:
            self.__class__.active_node = None
            instance = self.__class__(
                self.name,
                self.parameters_class(),
                self.description,
                machine=self.machine,
                modes=modes,
            )
            instance.modes.inspection = self.modes.inspection
            instance.filepath = self.filepath
            return instance
        finally:
            self.__class__.active_node = active_node

    def copy(self, name: str | None = None, **node_parameters: Any) -> Self:
        """
        Creates a modified copy of the node with updated parameters.

        The method allows the user to update parameters and assign a new name.
        If a new name is not provided, the copied node retains the same name as
        the original.

        Args:
            name: A new name for the node. Defaults to None.
            node_parameters: Additional parameters for the copied node.

        Returns:
            A copied node with the new parameters and name.

        Raises:
            ValueError: If the name provided is not a string.
        """
        logger.info(f"Copying node with name {self.name} with parameters {name = }, {node_parameters = }")
        if name is not None and not isinstance(name, str):
            raise ValueError(f"{self.__class__.__name__} should have a string name")
        instance = self.__copy__()
        if name is not None:
            instance.name = name
        instance._parameters = instance.parameters_class.model_validate(node_parameters)
        # Base class is inherited from user passed model so don't use passed
        # class as base for copied parameters class
        instance.parameters_class = self.build_parameters_class_from_instance(instance._parameters)
        return instance

    def set_parameters(self, **parameters: Any) -> None:
        self._parameters = self.parameters_class.model_validate(parameters)
        self.parameters_class = self.build_parameters_class_from_instance(self._parameters)

    def _warn_if_external_and_interactive_mpl(self) -> None:
        """
        Checks backend configuration and issues a warning if incompatible.

        Specifically, if the node is set to run in external mode and uses
        an interactive matplotlib backend, it switches to a non-interactive
        backend and logs a warning message.

        """
        mpl_backend = matplotlib.get_backend()
        if self.modes.external and mpl_backend in cast(
            list[str],
            backend_registry.list_builtin(BackendFilter.INTERACTIVE),  # type: ignore[no-untyped-call]
        ):
            plt.close("all")
            matplotlib.use("agg")
            logger.warning(
                f"Using interactive matplotlib backend '{mpl_backend}' in "
                "external mode. The backend is changed to 'agg'."
            )

    def __str__(self) -> str:
        return f"{self.__class__.__name__}: {self.name}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}: {self.name} (mode: {self.modes}; parameters: {self.parameters})"

    @property
    def snapshot_idx(self) -> int | None:
        """
        Returns the snapshot index from the storage manager.

        Retrieves the snapshot index that reflects the current state in
        the associated storage, if any.

        Returns:
            Snapshot index or None.
        """
        if self.storage_manager is None:
            return None
        return self.storage_manager.snapshot_idx

    def set_workflow_parent_id(self, parent_id: int | None) -> None:
        """Set the parent workflow ID for this node's snapshot.

        When running within a workflow, this ensures the node's snapshot is
        created with workflow_parent_id in its metadata from the start,
        preventing it from appearing at the top level of history.

        Args:
            parent_id: The snapshot ID of the parent workflow, or None.
        """
        storage_manager = self._get_storage_manager()
        if hasattr(storage_manager, "set_workflow_parent_id"):
            storage_manager.set_workflow_parent_id(parent_id)

    def run_action(
        self,
        func: ActionCallableType | None = None,
        *,
        skip_if: bool = False,
    ) -> ActionDecoratorType:
        """
        Convenience method that returns the run_action decorator
        provided by the ActionManager.

        Usage examples:

            @node.run_action
            def action1(node):
                # action code

            @node.run_action(skip_if=True)
            def action2(node):
                # action code; this action is skipped if
                # skip_if is True.
        """
        return self._action_manager.register_action(self, func, skip_if=skip_if)

    def _get_storage_manager(self) -> StorageManager[Self]:
        if self.storage_manager is not None:
            return self.storage_manager
        q_config_path = get_qualibrate_config_path()
        qs = get_qualibrate_config(q_config_path)
        state_path = get_quam_state_path(qs)
        self.storage_manager = LocalStorageManager[Self](
            root_data_folder=qs.storage.location,
            active_machine_path=state_path,
        )
        self.storage_manager.get_snapshot_idx(self)
        return self.storage_manager

    def save(self) -> None:
        """
        Saves the current state of the node to the storage manager.

        If no storage manager is assigned, the method attempts to create
        a default configuration for storage and logs a warning message.

        Raises:
            ImportError: Raised if required configurations are not accessible.
        """
        self._get_storage_manager().save(node=self)
        self.last_saved_at = datetime.now().astimezone()

    def _load_from_id(
        self,
        node_id: int,
        base_path: Path | None = None,
        custom_loaders: Sequence[type[BaseLoader]] | None = None,
        build_params_class: bool = False,
    ) -> Self | None:
        """
        Loads a node by its identifier, parsing its content and data.

        Args:
            node_id: The unique identifier of the node.
            base_path: The base directory where node data is stored. If None,
                attempts to retrieve the path from qualibrate app settings.
            custom_loaders: An optional sequence of custom loader classes
                for handling specific file types.
            build_params_class: Whether to dynamically build a parameters class
                based on the node schema.

        Returns:
            The current `QualibrationNode` instance with the loaded data,
            or None if loading fails.
        """
        if base_path is None:
            q_config_path = get_qualibrate_config_path()
            qs = get_qualibrate_config(q_config_path)
            base_path = qs.storage.location
        node_dir = get_node_dir_path(node_id, base_path)
        if node_dir is None:
            logger.error(f"Node directory with id {node_id} wasn't found in {base_path}")
            return None
        node_content = read_node_content(node_dir, node_id, base_path)
        if node_content is not None:
            quam_machine, parameters = parse_node_content(
                node_content,
                node_id,
                node_dir,
                build_params_class,
            )
            if quam_machine is not None:
                self.machine = quam_machine
            if parameters is not None:
                if build_params_class:
                    self.parameters_class = cast(ParametersType, parameters).__class__
                    self._parameters = cast(ParametersType, parameters)
                else:
                    self._parameters = cast(
                        ParametersType,
                        self.parameters.model_construct(**cast(Mapping[str, Any], parameters)),
                    )

        data = read_node_data(node_dir, node_id, base_path, custom_loaders)
        if data is not None:
            self.results = data
        return self

    @InstanceOrClassMethod
    def load_from_id(
        caller: Union[
            "QualibrationNode[ParametersType, NodeMachineType]",
            type["QualibrationNode[ParametersType, NodeMachineType]"],
        ],
        node_id: int,
        base_path: Path | None = None,
        custom_loaders: Sequence[type[BaseLoader]] | None = None,
    ) -> Optional["QualibrationNode[ParametersType, NodeMachineType]"]:
        """
        Class or instance method to load a node by its identifier.

        Args:
            caller: The class or instance calling this method. If called on
                a class, creates a new instance; otherwise, modifies the
                existing instance.
            node_id: The unique identifier of the node.
            base_path: The base directory where node data is stored. If None,
                attempts to retrieve the path from qualibrate app settings.
            custom_loaders: An optional sequence of custom loader classes
                for handling specific file types.

        Returns:
            A `QualibrationNode` instance with the loaded data, or None if
            loading fails.
        """
        instance: QualibrationNode[ParametersType, NodeMachineType] = (
            caller(name=f"loaded_from_id_{node_id}") if isinstance(caller, type) else caller
        )
        return instance._load_from_id(
            node_id=node_id,
            base_path=base_path,
            custom_loaders=custom_loaders,
            build_params_class=isinstance(caller, type),
        )

    def _extract_machines_metadata(self) -> Mapping[str, Any] | None:
        metadata = None
        try:
            if (
                self.parameters is not None
                and hasattr(self.parameters, "qubits")
                and self.machine is not None
                and hasattr(self.machine, "active_qubits")
                and hasattr(self.machine, "qubits")
            ):
                metadata = {}
                qubits = self.machine.qubits.keys()
                active_qubits = {qubit.name for qubit in self.machine.active_qubits}

                # Build metadata for each qubit
                for qubit in qubits:
                    qubit_info = self.machine.qubits[qubit]
                    gate_fidelity = None
                    if hasattr(qubit_info, "gate_fidelity") and hasattr(qubit_info.gate_fidelity, "averaged"):
                        gate_fidelity = qubit_info.gate_fidelity.averaged
                    metadata[qubit] = {
                        "active": qubit in active_qubits,
                        "fidelity": gate_fidelity,
                    }
        except Exception:
            # for any exception that could happen, return the old data as it was
            logger.info(
                "Failed to set quam machine metadata, probably used a different"
                " quam package hence the machine json load isn't compatible"
            )
            return None
        return metadata

    def _get_machine_metadata(self) -> Mapping[str, Any] | None:
        if self._machine is None:
            return None

        if self._machine_metadata is None:
            # Only extract metadata if machine has the right attributes
            try:
                self._machine_metadata = self._extract_machines_metadata()
            except AttributeError:
                self._machine_metadata = None

        # Always return dict (even empty) if machine has qubits
        return self._machine_metadata

    def serialize(self, **kwargs: Any) -> Mapping[str, Any]:
        # Get base QRunnable serialization
        data = dict(super().serialize(**kwargs))

        metadata = self._get_machine_metadata()

        if "parameters" in data and "qubits" in data["parameters"] and metadata is not None:
            data["parameters"]["qubits"]["metadata"] = metadata

        return data

    def _post_run(
        self,
        initial_targets: Sequence[TargetType],
        parameters: NodeParameters,
        run_error: RunError | None,
    ) -> NodeRunSummary:
        """
        Finalizes the node's execution and creates a summary.

        This method updates the outcomes for the last executed node,
        generates a summary, and logs the summary details. It is used
        to encapsulate the results of the node run, including the error
        (if any), state changes, and outcomes for each target.

        Args:
            last_executed_node: The node that was last executed.
            created_at: The timestamp when the run started.
            initial_targets: Targets at the start of the run.
            parameters: Parameters used in the run.
            run_error: Details of any error that occurred.

        Returns:
            A summary object containing execution details.
        """
        outcomes = self.outcomes
        self._action_manager.current_action = None
        if self.parameters is not None and (targets := self.parameters.targets):
            lost_targets_outcomes = set(targets) - set(outcomes.keys())
            outcomes.update({target: Outcome.SUCCESSFUL for target in lost_targets_outcomes})
        self.outcomes = {name: Outcome(outcome) for name, outcome in outcomes.items()}
        self.run_summary = NodeRunSummary(
            name=self.name,
            description=self.description,
            created_at=self.run_start,
            completed_at=datetime.now().astimezone(),
            initial_targets=initial_targets,
            parameters=parameters,
            outcomes=self.outcomes,
            error=run_error,
            successful_targets=[name for name, status in self.outcomes.items() if status == Outcome.SUCCESSFUL],
            failed_targets=[name for name, status in self.outcomes.items() if status == Outcome.FAILED],
            state_updates=self.state_updates,
        )
        logger.debug(f"Node run summary {self.run_summary}")
        return self.run_summary

    def _extract_source_snippet(self, tb: Any) -> str | None:
        """
        Extract source code snippet from the node file where error occurred.

        Returns the relevant lines from the node file with line numbers,
        or None if source cannot be read.

        Uses LAST occurrence strategy - finds the deepest point in the node file
        where the error occurred (could be in a subroutine, action, or body).

        Args:
            tb: The exception's traceback

        Returns:
            Formatted source snippet with line numbers and error marker,
            or None if source cannot be read
        """
        try:
            if not self.filepath:
                return None
            # Find the LAST frame in the node file
            # (deepest point where error occurred)
            # Note: This is different from simplify_traceback
            # - we want the actual error location
            all_frames = traceback.extract_tb(tb)
            node_filepath_str = str(self.filepath)

            node_frame_indices = [idx for idx, frame in enumerate(all_frames) if frame.filename == node_filepath_str]

            if not node_frame_indices:
                return None

            # Get the LAST occurrence
            # - this is where the error actually occurred
            # (could be in a subroutine, action function, or body code)
            error_frame = all_frames[node_frame_indices[-1]]
            error_lineno = error_frame.lineno
            if not error_lineno:
                return None

            # Read the source file
            with open(self.filepath) as f:
                lines = f.readlines()

            # Extract 2 lines before and 2 lines after (5 lines total)
            # -3 because lineno is 1-indexed
            start_line = max(0, error_lineno - 3)
            end_line = min(len(lines), error_lineno + 2)

            snippet_lines = []
            for i in range(start_line, end_line):
                line_num = i + 1
                line_text = lines[i].rstrip()
                marker = "  # <- Error occurred here" if line_num == error_lineno else ""
                snippet_lines.append(f"{line_num:4d}: {line_text}{marker}")

            return "\n".join(snippet_lines)
        except Exception:
            # If we can't read the source, just return None
            return None

    def _generate_error_headline(self, ex: Exception) -> str:
        """
        Generate a headline describing the error.

        Returns a string like:
        - "KeyError in action 'execute_qua_program'"
        - "ValueError in action 'process_data'"
        - "RuntimeError in node body"
        - "ImportError in node initialization"

        Args:
            ex: The exception that was raised

        Returns:
            Formatted headline string
        """
        failed_action = self._action_manager.failed_action

        if failed_action:
            return f"Failed action '{failed_action}'"
        else:
            # Determine if it's in node body or during initialization
            # If we have any actions registered, it's likely in the body
            # Otherwise, it's during initialization
            location = "node initialization (outside of actions)" if self._action_manager.actions else "node body"
            return f"Failure in {location}"

    def _generate_error_details(self, ex: Exception, simplified_tb: list[str]) -> str:
        """
        Generate detailed error context including action history,
        source snippet, and simplified traceback.

        Returns a simple markdown string with:
        - Failed action name (if applicable)
        - Source code snippet from node file
        - Simplified traceback
        - List of successfully completed actions
        - List of skipped actions
        - Error type and message

        Uses simple markdown compatible with React renderers
        (no nested code blocks or emojis).

        Args:
            ex: The exception that was raised
            simplified_tb: Simplified traceback (list of formatted strings)

        Returns:
            Formatted details string (simple markdown)
        """
        lines = []

        # Source code snippet (only from node file)
        source_snippet = self._extract_source_snippet(ex.__traceback__)
        if source_snippet:
            lines.append("Source Code:")
            # Add snippet lines with indentation
            for line in source_snippet.split("\n"):
                lines.append(f"  {line}")
            lines.append("")

        # Simplified traceback
        if simplified_tb:
            lines.append("Traceback:")
            # Add traceback lines with minimal indentation
            for tb_line in simplified_tb:
                # Remove extra whitespace and format consistently
                lines.append(f"  {tb_line.rstrip()}")
            lines.append("")

        # Action history (only if actions were used)
        completed = self._action_manager.completed_actions
        skipped = self._action_manager.skipped_actions

        if completed or skipped:
            if completed:
                lines.append("Completed actions (in order):")
                for action in completed:
                    lines.append(f"  - {action}")
                lines.append("")

            if skipped:
                lines.append("Skipped actions:")
                for action in skipped:
                    lines.append(f"  - {action}")
                lines.append("")

        return "\n".join(lines)

    def run(
        self,
        interactive: bool = False,
        *,
        skip_actions: bool | Sequence[str] = False,
        **passed_parameters: Any,
    ) -> BaseRunSummary:
        """
        Runs the node with given parameters, potentially interactively.

        This function executes the node using parameters passed either
        directly or pre-configured in the node. It captures initial
        conditions, manages execution state, and logs errors or completion
        status.

        Args:
            interactive: Whether the node should be run interactively.
            skip_actions: An optional sequence of actions to skip when running.
                Also, possible to pass `True` to skip all actions.
            **passed_parameters: Additional parameters to pass when
                running the node.

        Returns:
            The executed node and a summary of the run including outcomes,
            errors, and execution details.

        Raises:
            RuntimeError: Raised if the node filepath is not provided, or
                execution
        """
        logger.info(f"Run node {self.name} with parameters: {passed_parameters}")
        self._fraction_complete = 0
        if self.filepath is None:
            ex = RuntimeError(f"Node {self.name} file path was not provided")
            logger.exception("", exc_info=ex)
            raise ex
        params_dict = self.parameters.model_dump() if self.parameters is not None else {}
        params_dict.update(passed_parameters)
        parameters = self.parameters.model_validate(params_dict)
        initial_targets = copy.copy(parameters.targets) if parameters.targets else []
        self.run_start = datetime.now().astimezone()
        run_error: RunError | None = None

        if run_modes_ctx.get() is not None:
            logger.error(
                "Run modes context is already set to %s",
                run_modes_ctx.get(),
            )
        new_run_modes = RunModes(external=True, interactive=interactive, inspection=False)
        run_modes_token = run_modes_ctx.set(new_run_modes)
        modes = self.modes.model_copy()
        try:
            self._parameters = parameters
            self.modes = new_run_modes
            self.__class__.active_node = self
            self._action_manager.skip_actions = skip_actions

            # Reset action execution history for this run
            self._action_manager.completed_actions.clear()
            self._action_manager.skipped_actions.clear()
            self._action_manager.failed_action = None

            self.run_node_file(self.filepath)
        except Exception as ex:
            # Generate simplified traceback for error details (user-friendly)
            simplified_tb = simplify_traceback(ex.__traceback__, self.filepath)

            # This particular RunError is propagated to the frontend
            run_error = RunError(
                error_class=ex.__class__.__name__,
                message=str(ex),
                details_headline=self._generate_error_headline(ex),
                details=self._generate_error_details(ex, simplified_tb),
                # Keep FULL traceback
                traceback=traceback.format_tb(ex.__traceback__),
            )
            logger.exception(f"Failed to run node {self.name}", exc_info=ex)
            raise
        else:
            self._fraction_complete = 1.0
        finally:
            self._action_manager.skip_actions = False
            run_modes_ctx.reset(run_modes_token)
            self.modes = modes
            self.__class__.active_node = None
            run_summary = self._post_run(
                initial_targets,
                parameters,
                run_error,
            )

        return run_summary

    def run_node_file(self, node_filepath: Path) -> None:
        """
        Executes the provided node file.

        This method runs the code in the given node file, ensuring that
        any interactive backends are temporarily disabled during the
        execution to avoid conflicts. Once the file has been executed,
        the original matplotlib backend is restored.

        Args:
            node_filepath: Path to the file that contains the node's execution
                logic.
        """
        mpl_backend = matplotlib.get_backend()
        # Appending dir with nodes can cause issues with relative imports
        try:
            plt.close("all")
            matplotlib.use("agg")
            _module = import_from_path(get_module_name(node_filepath), node_filepath)
        finally:
            matplotlib.use(mpl_backend)

    def stop(self, **kwargs: Any) -> bool:
        """
        Halts the execution of the node if currently running.

        The method attempts to get the qm job from namespace and halt
        got job. Also ship all subsequent run actions.

        Args:
            **kwargs: Additional keyword arguments that might be used for
                stopping the node.

        Returns:
            True if the node is successfully stopped, False otherwise.
        """
        active_node = self.__class__.active_node
        if active_node is None:
            logger.warning("No active node to stop")
            return False
        logger.info(f"Trying to stop node {active_node.name}")
        job = self.namespace.get("job")
        if job is not None and callable(getattr(job, "halt", None)):
            job.halt()
        active_node._action_manager.skip_actions = True
        return True

    @contextmanager
    def record_state_updates(self, interactive_only: bool = True) -> Generator[None, None, None]:
        """
        Records state updates for the node during execution.

        This method wraps around node operations to record all state updates
        that occur, specifically during interactive execution. It uses
        custom `__setattr__` and `__setitem__` functions for relevant classes
        to record these changes.
        Only simple types updates (int, float, bool, str and None) will
        be recorded.

        Args:
            interactive_only: Whether to only record in interactive mode.
                Defaults to True.

        Yields:
            None: Allows wrapped operations to execute while recording state
                updates.
        """
        machine = self.machine
        if not self.modes.interactive and interactive_only:
            yield
            return

        if machine is None or not hasattr(machine, "to_dict"):
            self.log(
                (
                    "Unable to perform `QualibrationNode."
                    "record_state_updates()` because `node.machine` has not "
                    "been set. Any changes will be automatically applied."
                ),
                level="warning",
            )
            yield
            return

        logger.debug(f"Init recording state updates for node {self.name}")
        original_dict = machine.to_dict(include_defaults=True)
        yield
        updated_dict = machine.to_dict(include_defaults=True)
        update_node_machine(self, original_dict, updated_dict)

    @classmethod
    def scan_folder_for_instances(
        cls, path: Path
    ) -> RunnableCollection[str, QRunnable[ParametersType, ParametersType]]:
        """
        Scans a directory for node instances and returns them.

        This method scans a folder to locate all node files that are valid
        instances of `QualibrationNode`. It sets an inspection mode to avoid
        executing the nodes during scanning.

        Args:
            path: The directory to scan for node files.

        Returns:
            A dictionary of node names to their corresponding node instances.
        """
        nodes: dict[str, QRunnable[ParametersType, ParametersType]] = {}
        if run_modes_ctx.get() is not None:
            logger.error(
                "Run modes context is already set to %s",
                run_modes_ctx.get(),
            )
        run_modes_token = run_modes_ctx.set(RunModes(inspection=True))
        try:
            for file in sorted(path.iterdir()):
                if not file_is_calibration_node_instance(file, cls.__name__):
                    continue
                try:
                    cls.scan_node_file(file, nodes)
                except Exception as e:
                    logger.warning(f"An error occurred on scanning node file {file.name}.\nError: {type(e)}: {e}")

        finally:
            run_modes_ctx.reset(run_modes_token)
        return RunnableCollection(nodes)

    @classmethod
    def scan_node_file(
        cls,
        file: Path,
        nodes: dict[str, QRunnable[ParametersType, ParametersType]],
    ) -> None:
        """
        Scans a node file and adds its instance to the provided dictionary.

        This method scans the content of a given file to identify if it
        contains a valid `QualibrationNode`. If so, it adds the node to
        the given nodes dictionary for further processing.

        Args:
            file: The node file to scan.
            nodes: dictionary to add valid nodes to.

        Raises:
            StopInspection: Used to stop execution once inspection completes.
        """
        logger.info(f"Scanning node file {file}")
        try:
            # TODO Think of a safer way to execute the code
            _module = import_from_path(get_module_name(file), file)
        except StopInspection as ex:
            node = cast("QualibrationNode[ParametersType, Any]", ex.instance)
            node.filepath = file
            node.modes.inspection = False
            cls.add_node(node, nodes)

    @classmethod
    def add_node(
        cls,
        node: "QualibrationNode[ParametersType, NodeMachineType]",
        nodes: dict[str, QRunnable[ParametersType, ParametersType]],
    ) -> None:
        """
        Adds a node instance to the node dictionary.

        If a node with the same name already exists in the dictionary,
        this method overwrites the existing entry with a warning.

        Args:
            node: The node instance to add.
            nodes: dictionary to store nodes.
        """
        if node.name in nodes:
            logger.warning(f'Node "{node.name}" already exists in library, overwriting')

        nodes[node.name] = node

    def log(
        self,
        msg: object,
        *args: Any,
        level: LOG_LEVEL_NAMES_TYPE | int = "info",
        **kwargs: Any,
    ) -> None:
        name = self.name if len(self.name) <= 20 else f"{self.name[:17]}..."
        new_message = f"Node {name} - {msg}"
        if isinstance(level, int):
            logger.log(level, new_message, *args, **kwargs)
            return
        if level not in ALLOWED_LOG_LEVEL_NAMES:
            raise ValueError("Invalid log level name. Can't write log message.")
        getattr(logger, level)(new_message, *args, **kwargs)

    @property
    def fraction_complete(self) -> float:
        return self._fraction_complete

    @fraction_complete.setter
    def fraction_complete(self, value: float) -> None:
        self._fraction_complete = max(min(value, 1.0), 0.0)

    @property
    def current_action_name(self) -> str | None:
        action = self._action_manager.current_action
        return action.name if action else None

    def cleanup(self) -> None:
        super().cleanup()
        self._fraction_complete = 0.0
        self.results = {}
        self.machine = None
        self.storage_manager = None
        self._action_manager = ActionManager()
        self.namespace = {}
        self._post_init()


if __name__ == "__main__":
    from pathlib import Path

    # path = Path("/home/maxim_v4s/Downloads/")
    path = Path.home().joinpath(".qualibrate/user_storage/init_project")
    node = QualibrationNode.load_from_id(12)
    print(node.results)
    print(node.parameters)



"""Utility class for handling snapshot JSON file operations.

This module provides a reusable handler for reading and writing
node.json files in the local storage format.
"""

import json
from pathlib import Path
from typing import TYPE_CHECKING

from qualibrate.core.utils.logger_m import logger

if TYPE_CHECKING:
    from typing import Any

__all__ = ["SnapshotJsonHandler"]


class SnapshotJsonHandler:
    """Handles JSON file operations for snapshots.

    Provides methods for finding, reading, and writing node.json files
    in the local storage directory structure.

    Args:
        root_data_folder: The root folder where snapshot data is stored.
    """

    def __init__(self, root_data_folder: Path):
        self.root_data_folder = root_data_folder

    def get_node_json_path(self, snapshot_idx: int) -> Path | None:
        """Get the path to a snapshot's node.json file by snapshot ID.

        Searches for a snapshot directory matching the ID pattern
        in the root data folder.

        Args:
            snapshot_idx: The snapshot ID to find.

        Returns:
            Path to the node.json file, or None if not found.
        """
        # Search for the snapshot directory using the ID pattern
        pattern = f"*/#{snapshot_idx}_*"
        matches = list(self.root_data_folder.glob(pattern))
        if not matches:
            logger.warning(f"Snapshot {snapshot_idx} not found in {self.root_data_folder}")
            return None
        return matches[0] / "node.json"

    def read_node_json(self, node_json_path: Path) -> dict[str, "Any"] | None:
        """Read and parse a node.json file.

        Args:
            node_json_path: Path to the node.json file.

        Returns:
            Parsed JSON content as a dictionary, or None if read fails.
        """
        if not node_json_path.is_file():
            logger.warning(f"node.json not found at {node_json_path}")
            return None
        try:
            with node_json_path.open("r") as f:
                return dict(json.load(f))
        except json.JSONDecodeError as ex:
            logger.exception(f"Failed to parse {node_json_path}", exc_info=ex)
            return None

    def write_node_json(self, node_json_path: Path, content: dict[str, "Any"]) -> bool:
        """Write content to a node.json file.

        Args:
            node_json_path: Path to the node.json file.
            content: Content to write as JSON.

        Returns:
            True if write succeeded, False otherwise.
        """
        try:
            with node_json_path.open("w") as f:
                json.dump(content, f, indent=2, default=str)
            return True
        except Exception as ex:
            logger.exception(f"Failed to write {node_json_path}", exc_info=ex)
            return False

    def read_snapshot(self, snapshot_idx: int) -> dict[str, "Any"] | None:
        """Read a snapshot's node.json content by its ID.

        Convenience method that combines get_node_json_path and read_node_json.

        Args:
            snapshot_idx: The snapshot ID to read.

        Returns:
            Parsed JSON content, or None if not found or read fails.
        """
        path = self.get_node_json_path(snapshot_idx)
        if path is None:
            return None
        return self.read_node_json(path)

    def write_snapshot(self, snapshot_idx: int, content: dict[str, "Any"]) -> bool:
        """Write content to a snapshot's node.json file by its ID.

        Convenience method that combines get_node_json_path and write_node_json.

        Args:
            snapshot_idx: The snapshot ID to write to.
            content: Content to write as JSON.

        Returns:
            True if write succeeded, False otherwise.
        """
        path = self.get_node_json_path(snapshot_idx)
        if path is None:
            return False
        return self.write_node_json(path, content)



from qualibrate_config.models import QualibrateConfig

from qualibrate.app.api.core.domain.bases.node import NodeBase
from qualibrate.app.api.core.domain.timeline_db.snapshot import (
    SnapshotTimelineDb,
)
from qualibrate.app.api.core.types import DocumentType, IdType

__all__ = ["NodeTimelineDb"]


class NodeTimelineDb(NodeBase):
    def __init__(
        self,
        node_id: IdType,
        snapshot_content: DocumentType | None = None,
        *,
        settings: QualibrateConfig,
    ):
        super().__init__(
            node_id,
            SnapshotTimelineDb(node_id, snapshot_content, settings=settings),
            settings,
        )



from qualibrate.app.api.exceptions.classes.base import QualibrateException


class QPathException(QualibrateException, OSError):
    pass


class QFileNotFoundException(QPathException, FileNotFoundError):
    pass


class QRelativeNotSubpathException(QPathException):
    pass


class QNotADirectoryException(QPathException, NotADirectoryError):
    pass



"""
Global state tracking for the qualibrate runner.

This module defines the State model, which tracks the currently executing
calibration job and its status. The State is used by:
- The web UI to display real-time execution status
- The API to report job progress
- Monitoring systems to track calibration workflows

The State is a singleton-like object (one per runner instance) that is
continuously updated as jobs execute.
"""

from pydantic import BaseModel, ConfigDict

from qualibrate.runner.core.models.enums import RunStatusEnum
from qualibrate.runner.core.models.last_run import LastRun
from qualibrate.runner.core.types import QGraphType, QNodeType

__all__ = ["State", "RunStatusEnum"]


class State(BaseModel):
    """
    Tracks the current and last executed calibration job.

    This model serves as the single source of truth for the runner's execution
    state. It is continuously updated by run_job.run_node() and
    run_job.run_workflow() as jobs execute.

    The State is designed to be serializable (via Pydantic) so it can be
    exposed through APIs for monitoring and UI purposes.

    Attributes:
        last_run: Complete information about the most recent execution,
            including status, timing, results, and errors. None if no job
            has been executed yet.
        run_item: Reference to the actual QualibrationNode or Graph being
            executed. This allows access to the live object during execution.
            None when no job is running.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    last_run: LastRun | None = None
    run_item: QNodeType | QGraphType | None = None

    @property
    def is_running(self) -> bool:
        return self.last_run is not None and self.last_run.status == RunStatusEnum.RUNNING

    def clear(self) -> None:
        """
        Clear the state, removing all execution history.

        This resets both last_run and run_item to None, effectively clearing
        all execution state. Used to start fresh or clean up after errors.

        Raises:
            RuntimeError: If called while a job is currently running (status
                is RUNNING). You must wait for the job to complete or error
                before clearing state.
        """
        if self.last_run is not None and self.last_run.status == RunStatusEnum.RUNNING:
            raise RuntimeError("Can't clear while item is running")
        self.last_run = None
        self.run_item = None



from collections import UserDict, UserList
from types import NoneType
from typing import TYPE_CHECKING, Any, Optional, TypeVar

import jsonpatch
import jsonpointer

from qualibrate.core.parameters import NodeParameters
from qualibrate.core.utils.logger_m import logger
from qualibrate.core.utils.type_protocols import MachineProtocol

if TYPE_CHECKING:
    from qualibrate.core.qualibration_node import QualibrationNode

try:
    from quam.utils import string_reference
except ModuleNotFoundError:
    string_reference = None


__all__ = [
    "record_state_update",
    "update_machine_attribute",
    "update_node_machine",
]

ParametersType = TypeVar("ParametersType", bound=NodeParameters)
MachineType = TypeVar("MachineType", bound=MachineProtocol)
ValueTypeToRecord = (int, float, str, NoneType)


def _record_state_update(
    node: Optional["QualibrationNode[ParametersType, MachineType]"],
    reference: str,
    attr: str,
    old: Any,
    val: Any,
) -> None:
    """
    Records state updates for an attribute or item in the node.

    This function stores information about changes made to an attribute
    or item of a node, including the previous value and the new value.
    If the node is provided, the change details are saved in the node's
    `_state_updates` dictionary.

    Args:
        node: The node where the state update will be recorded. If None, no
            action is performed.
        reference: The reference key to identify the updated attribute or item.
        attr: The name of the attribute or item key that is updated.
        old: The old value of the attribute or item before the update.
        val: The new value of the attribute or item.
    """
    if node is None:
        return
    if isinstance(old, UserList):
        old = list(old)
    elif isinstance(old, UserDict):
        old = dict(old)
    node._state_updates[reference] = {
        "key": reference,
        "attr": attr,
        "old": old,
        "new": val,
    }


def record_state_update(
    node: "QualibrationNode[ParametersType, MachineType]",
    reference: str,
    attr: str,
    old: Any,
    val: Any = None,
) -> None:
    """
    Records item state updates in a Quam collection object.

    For details see `_record_state_update`.

    Args:
        node: The node where the state update will be recorded. Defaults to
            None.
        attr: The name of the attribute being updated.
        old: The old value of the attribute or item.
        val: The new value of the attribute or item. Defaults to None.
    """
    _record_state_update(node, reference, attr, old, val)


def update_machine_attribute(
    machine: MachineType,
    quam_path: str,
    old_value: Any,
) -> str | int:
    if string_reference is None:
        raise RuntimeError("QUAM is not installed, skipping state update recording")
    ref_path, str_key = string_reference.split_reference(quam_path)
    try:
        key = int(str_key)
    except ValueError:
        key = str_key
    referenced_value = string_reference.get_referenced_value(machine, ref_path, root=machine.get_root())
    if isinstance(referenced_value, (list, UserList)) and isinstance(key, int):
        referenced_value[key] = old_value
        return key
    if isinstance(referenced_value, (dict, UserDict)):
        if str_key in referenced_value:
            referenced_value[str_key] = old_value
            return str(str_key)
        if key in referenced_value:
            referenced_value[key] = old_value
            return key
    setattr(referenced_value, str_key, old_value)
    return key


def update_node_machine(
    node: "QualibrationNode[ParametersType, MachineType]",
    original_dict: dict[str, Any],
    updated_dict: dict[str, Any],
) -> None:
    if node.machine is None:
        return
    if string_reference is None:
        logger.warning("QUAM is not installed, skipping state update recording")
        return
    patches: jsonpatch.JsonPatch = jsonpatch.make_patch(original_dict, updated_dict)
    for patch in patches.patch:
        if patch["op"] != "replace" or not isinstance(patch["value"], ValueTypeToRecord):
            continue
        path_ptr = jsonpointer.JsonPointer(patch["path"])
        try:
            old_value = jsonpointer.resolve_pointer(original_dict, path_ptr.path)
        except jsonpointer.JsonPointerException as ex:
            logger.exception(ex)
            continue
        if (
            string_reference.is_reference(old_value)
            or string_reference.is_reference(patch["value"])
            or not isinstance(old_value, ValueTypeToRecord)
        ):
            continue
        quam_path = f"#{path_ptr.path}"
        attr = update_machine_attribute(node.machine, quam_path, old_value)
        record_state_update(
            node=node,
            reference=quam_path,
            attr=str(attr),
            old=old_value,
            val=patch["value"],
        )



"""API routes for global tag management."""

from typing import Annotated

from fastapi import APIRouter, Body, Depends, status
from qualibrate_config.models import QualibrateConfig

from qualibrate.app.api.core.domain.local_storage.tag_registry import TagRegistry
from qualibrate.app.api.core.schemas.tag import TagNameRequest
from qualibrate.app.config import get_settings

__all__ = ["tag_router"]

tag_router = APIRouter(prefix="/snapshot", tags=["tags"])


def _get_tag_registry(
    settings: Annotated[QualibrateConfig, Depends(get_settings)],
) -> TagRegistry:
    """Get the tag registry instance for the current project."""
    return TagRegistry(settings=settings)


@tag_router.post(
    "/tag/create",
    summary="Create a new global tag",
    response_model=bool,
    responses={
        status.HTTP_200_OK: {
            "description": "Tag creation result",
            "content": {
                "application/json": {
                    "examples": {
                        "success": {
                            "summary": "Tag created successfully",
                            "value": True,
                        },
                        "failure": {
                            "summary": "Tag creation failed",
                            "value": False,
                        },
                    }
                }
            },
        }
    },
)
def create_tag(
    request: Annotated[TagNameRequest, Body()],
    tag_registry: Annotated[TagRegistry, Depends(_get_tag_registry)],
) -> bool:
    """
    Create a new global tag in the tag registry.

    If the tag already exists, returns True without creating a duplicate.
    Tags are stored in a project-scoped `tags.json` file.

    ### Example

    **Request:**
    ```json
    {"name": "calibration"}
    ```

    **Response:** `true` or `false`
    """
    return tag_registry.create_tag(request.name)


@tag_router.get(
    "/tags",
    summary="List all global tags",
    response_model=list[str],
    responses={
        status.HTTP_200_OK: {
            "description": "List of all registered tags",
            "content": {
                "application/json": {
                    "examples": {
                        "with_tags": {
                            "summary": "Tags exist",
                            "value": ["benchmarking", "calibration", "rabi"],
                        },
                        "empty": {
                            "summary": "No tags",
                            "value": [],
                        },
                    }
                }
            },
        }
    },
)
def list_tags(
    tag_registry: Annotated[TagRegistry, Depends(_get_tag_registry)],
) -> list[str]:
    """
    Get all registered global tags.

    Returns a sorted list of all tag names in the project's tag registry.

    ### Example

    **Response:**
    ```json
    ["benchmarking", "calibration", "quick-check", "rabi"]
    ```
    """
    return tag_registry.list_tags()


@tag_router.post(
    "/tag/remove",
    summary="Remove a global tag",
    response_model=bool,
    responses={
        status.HTTP_200_OK: {
            "description": "Tag removal result",
            "content": {
                "application/json": {
                    "examples": {
                        "success": {
                            "summary": "Tag removed successfully",
                            "value": True,
                        },
                        "failure": {
                            "summary": "Tag removal failed",
                            "value": False,
                        },
                    }
                }
            },
        }
    },
)
def remove_tag(
    request: Annotated[TagNameRequest, Body()],
    tag_registry: Annotated[TagRegistry, Depends(_get_tag_registry)],
) -> bool:
    """
    Remove a global tag from the tag registry.

    **Note:** This only removes the tag from the global registry.
    Existing snapshots that have this tag assigned will retain it
    (soft delete behavior).

    If the tag doesn't exist, returns True.

    ### Example

    **Request:**
    ```json
    {"name": "calibration"}
    ```

    **Response:** `true` or `false`
    """
    return tag_registry.remove_tag(request.name)



"""Request/Response schemas for tag-related endpoints."""

from pydantic import BaseModel, Field

__all__ = [
    "TagNameRequest",
    "TagsAssignRequest",
]


class TagNameRequest(BaseModel):
    """Request body for tag create/remove operations."""

    name: str = Field(
        ...,
        min_length=1,
        description="The tag name.",
        examples=["calibration", "rabi", "benchmarking"],
    )


class TagsAssignRequest(BaseModel):
    """Request body for assigning multiple tags to a snapshot."""

    tags: list[str] = Field(
        ...,
        description="List of tag names to assign. Use empty list to clear all tags.",
        examples=[["calibration", "rabi"], ["quick-check"], []],
    )



from fastapi import APIRouter

from qualibrate.app.api.routes import (
    other_router,
    project_router,
    projects_router,
    storage_router,
)

api_router = APIRouter()

api_router.include_router(other_router)
api_router.include_router(project_router)
api_router.include_router(projects_router)
api_router.include_router(storage_router)



from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from urllib.parse import urljoin

import requests

from qualibrate.app.api.core.domain.bases.project import ProjectsManagerBase
from qualibrate.app.api.core.models.project import Project
from qualibrate.app.api.core.utils.request_utils import request_with_db
from qualibrate.app.api.exceptions.classes.timeline_db import QJsonDbException
from qualibrate.app.api.exceptions.classes.values import QValueException


class ProjectsManagerTimelineDb(ProjectsManagerBase):
    def _check_db_project_exists(self, project_name: str) -> bool:
        response = request_with_db(
            "database/connect",
            db_name=project_name,
            host=self._settings.timeline_db.address_with_root,
            timeout=self._settings.timeline_db.timeout,
        )
        return response.status_code == 200 and response.json() is True

    def _active_project_setter(self, value: str) -> None:
        if not self._check_db_project_exists(value):
            raise QJsonDbException(f"Can't check if project {value} exists in timeline DB.")
        self._settings.project = value

    def create(
        self,
        project_name: str,
        storage_location: Path | None = None,
        calibration_library_folder: Path | None = None,
        quam_state_path: Path | None = None,
    ) -> str:
        if any(project.name == project_name for project in self.list()):
            raise QValueException(f"Project {project_name} already exists.")
        response = request_with_db(
            "database/create",
            db_name=project_name,
            host=self._settings.timeline_db.address_with_root,
            timeout=self._settings.timeline_db.timeout,
            method=requests.post,
        )

        if response.status_code != 200 or response.json() != project_name:
            raise QJsonDbException(f"Can't create project {project_name}.")
        return super().create(
            project_name,
            storage_location,
            calibration_library_folder,
            quam_state_path,
        )

    def list(self) -> Sequence[Project]:
        response = requests.get(
            urljoin(self._settings.timeline_db.address_with_root, "database/list"),
            timeout=self._settings.timeline_db.timeout,
        )
        if response.status_code != 200:
            raise QJsonDbException("Can't resolve list of projects.")
        return list(
            Project(
                name=name,
                nodes_number=-1,
                created_at=datetime.fromtimestamp(0),
                last_modified_at=datetime.fromtimestamp(0),
            )
            for name in response.json()
        )



from collections.abc import Sequence
from datetime import datetime
from pathlib import Path

from qualibrate_config.models import QualibrateConfig

from qualibrate.app.api.core.domain.bases.branch import (
    BranchBase,
    BranchLoadType,
)
from qualibrate.app.api.core.domain.bases.node import NodeBase, NodeLoadType
from qualibrate.app.api.core.domain.bases.snapshot import (
    SnapshotBase,
    SnapshotLoadTypeFlag,
)
from qualibrate.app.api.core.domain.local_storage.node import NodeLocalStorage
from qualibrate.app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
)
from qualibrate.app.api.core.domain.local_storage.utils.local_path_id import (
    IdToLocalPath,
)
from qualibrate.app.api.core.domain.local_storage.utils.node_utils import (
    find_nodes_ids_by_filter,
)
from qualibrate.app.api.core.models.branch import Branch as BranchModel
from qualibrate.app.api.core.models.snapshot import SnapshotSearchResult
from qualibrate.app.api.core.types import (
    DocumentType,
    IdType,
    PageFilter,
    SearchWithIdFilter,
)
from qualibrate.app.api.core.utils import find_utils
from qualibrate.app.api.core.utils.slice import get_page_slice
from qualibrate.app.api.exceptions.classes.storage import QFileNotFoundException

__all__ = ["BranchLocalStorage"]


class BranchLocalStorage(BranchBase):
    def __init__(
        self,
        name: str,
        content: DocumentType | None = None,
        *,
        settings: QualibrateConfig,
    ):
        # Temporary branch name has no effect
        super().__init__(name, content, settings=settings)

    @property
    def created_at(self) -> datetime:
        return datetime.fromtimestamp(self._settings.storage.location.stat().st_mtime).astimezone()

    def load(self, load_type: BranchLoadType) -> None:
        pass

    def _get_latest_node_id(self, error_msg: str) -> IdType:
        node_id = next(
            find_nodes_ids_by_filter(
                self._settings.storage.location,
                project_name=self._settings.project,
                descending=True,
            ),
            None,
        )
        if node_id is None:
            raise QFileNotFoundException(f"There is no {error_msg}")
        return node_id

    def get_snapshot(self, snapshot_id: IdType | None = None) -> SnapshotBase:
        if snapshot_id is None:
            snapshot_id = self._get_latest_node_id("snapshots")
        return SnapshotLocalStorage(snapshot_id, settings=self._settings)

    def get_node(self, id: IdType | None = None) -> NodeBase:
        if id is None:
            id = self._get_latest_node_id("nodes")
        return NodeLocalStorage(id, settings=self._settings)

    def _get_filtered_ids(
        self,
        storage_location: Path,
        search_filter: SearchWithIdFilter | None = None,
    ) -> set[IdType]:
        """Get filtered IDs without pagination or sorting.

        This is used to calculate the total count of filtered items.
        """
        storage = IdToLocalPath().get_project_manager(self._settings.project, storage_location)
        return storage.get_ids(search_filter)

    def _get_latest_snapshots_ids(
        self,
        storage_location: Path,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        descending: bool = False,
    ) -> Sequence[IdType]:
        ids = find_nodes_ids_by_filter(
            storage_location,
            search_filter=search_filter,
            project_name=self._settings.project,
            descending=descending,
        )
        return get_page_slice(ids, pages_filter)

    def get_latest_snapshots(
        self,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        descending: bool = False,
        include_outcomes: bool = False,
    ) -> tuple[int, Sequence[SnapshotBase]]:
        storage_location = self._settings.storage.location
        # Get total count of filtered items (before pagination)
        filtered_ids = self._get_filtered_ids(storage_location, search_filter)
        total = len(filtered_ids)

        ids_paged = self._get_latest_snapshots_ids(
            storage_location,
            pages_filter=pages_filter,
            search_filter=search_filter,
            descending=descending,
        )
        snapshots = [SnapshotLocalStorage(id, settings=self._settings) for id in ids_paged]
        # Load metadata, and optionally outcomes data for workflow aggregation
        load_flag = SnapshotLoadTypeFlag.Metadata
        if include_outcomes:
            load_flag = load_flag | SnapshotLoadTypeFlag.DataWithoutRefs
        for snapshot in snapshots:
            snapshot.load_from_flag(load_flag)
        return total, snapshots

    def get_latest_nodes(
        self,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        descending: bool = False,
    ) -> tuple[int, Sequence[NodeBase]]:
        storage_location = self._settings.storage.location
        # Get total count of filtered items (before pagination)
        filtered_ids = self._get_filtered_ids(storage_location, search_filter)
        total = len(filtered_ids)

        ids_paged = self._get_latest_snapshots_ids(
            storage_location,
            pages_filter=pages_filter,
            search_filter=search_filter,
            descending=descending,
        )
        nodes = [NodeLocalStorage(id, settings=self._settings) for id in ids_paged]
        for node in nodes:
            node.load(NodeLoadType.Full)

        return total, nodes

    def dump(self) -> BranchModel:
        return BranchModel(
            id=1,
            created_at=self.created_at,
            name=self._name,
            snapshot_id=-1,
        )

    def search_snapshots_data(
        self,
        *,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        data_path: Sequence[str | int],
        filter_no_change: bool = True,
        descending: bool = False,
    ) -> tuple[int, Sequence[SnapshotSearchResult]]:
        storage_location = self._settings.storage.location
        ids = find_nodes_ids_by_filter(
            storage_location,
            search_filter=search_filter,
            project_name=self._settings.project,
            descending=descending,
        )
        snapshots = (SnapshotLocalStorage(id, settings=self._settings) for id in ids)
        if descending:
            snapshots_with_data = find_utils.search_snapshots_data_with_filter_descending(
                snapshots, data_path, filter_no_change
            )
        else:
            snapshots_with_data = find_utils.search_snapshots_data_with_filter_ascending(
                snapshots, data_path, filter_no_change
            )
        return 0, get_page_slice(snapshots_with_data, pages_filter)



from collections.abc import Callable, Generator
from typing import Generic

from pydantic import BaseModel

from qualibrate.core.utils.graph_building import GraphElementTypeVar
from qualibrate.core.utils.type_protocols import TargetType


class OperationalCondition(BaseModel, Generic[GraphElementTypeVar]):
    on_function: Callable[[GraphElementTypeVar, TargetType], bool] | None = None
    on_generator: (
        Callable[
            [],
            Generator[bool, tuple[GraphElementTypeVar, TargetType] | None, None],
        ]
        | None
    ) = None


class LoopCondition(OperationalCondition[GraphElementTypeVar]):
    max_iterations: int | None = None
    on_failure: bool = False



from importlib.util import find_spec
from pathlib import Path
from typing import Any

from qualibrate.core.utils.node.loaders import BaseLoader


class XarrayLoader(BaseLoader):
    """
    Loader for xarray datasets.

    Attributes:
        file_extensions: A tuple of supported file extensions (e.g., ".h5").
    """

    file_extensions: tuple[str, ...] = (".h5",)

    def load(self, path: Path, **kwargs: Any) -> Any:
        """
        Loads a xarray dataset from a file.

        Args:
            path: The path to the xarray file.
            **kwargs: Additional arguments for file loading.

        Returns:
            The loaded xarray dataset, or None if the xarray library is
            unavailable.
        """
        if find_spec("xarray") is None:
            return None
        from xarray import open_dataset

        self.__class__.validate_file_exists(path)
        return open_dataset(path)



"""Demo: T1 Measurement - simulates energy relaxation time."""

# %% Imports
import time
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit

from qualibrate.core import NodeParameters, QualibrationNode


# %% Parameters
class Parameters(NodeParameters):
    """Parameters for demo T1 measurement."""

    qubits: list[str] = ["q1", "q2"]
    num_shots: int = 100
    max_delay: float = 50e-6  # seconds (50 microseconds)
    num_points: int = 40
    duration: float = 10.0  # seconds


# %% Helper Functions
def _exponential_decay(
    delay: np.ndarray, t1: float, amplitude: float, offset: float
) -> np.ndarray:
    """Exponential decay for T1 relaxation."""
    return amplitude * np.exp(-delay / t1) + offset


def _generate_t1_data(delays: np.ndarray, noise_level: float = 0.03) -> np.ndarray:
    """Generate realistic T1 exponential decay with noise."""
    t1 = np.random.uniform(15e-6, 25e-6)  # T1 time (seconds)
    amplitude = np.random.uniform(0.85, 0.95)  # Initial population
    offset = np.random.uniform(0.0, 0.05)  # Residual population
    signal = _exponential_decay(delays, t1, amplitude, offset)
    noise = np.random.normal(0, noise_level, size=len(delays))
    return np.clip(signal + noise, 0, 1)


def _fit_t1(delays: np.ndarray, populations: np.ndarray) -> dict:
    """Fit exponential decay to extract T1."""
    try:
        # Initial guesses
        amplitude_guess = populations[0] - populations[-1]
        offset_guess = populations[-1]
        t1_guess = delays[len(delays) // 2]

        popt, _ = curve_fit(
            _exponential_decay,
            delays,
            populations,
            p0=[t1_guess, amplitude_guess, offset_guess],
            bounds=([1e-6, 0.1, -0.1], [100e-6, 1.5, 0.5]),
            maxfev=10000,
        )
        return {
            "success": True,
            "t1": popt[0],
            "amplitude": popt[1],
            "offset": popt[2],
            "fit_params": popt,
        }
    except Exception as e:
        return {"success": False, "error": str(e)}


# %% Node Definition
description = """
DEMO: T1 MEASUREMENT (Simulated)

Simulates measurement of the energy relaxation time (T1) by preparing
the qubit in the excited state and measuring population decay as a
function of wait time. T1 sets the ultimate limit on coherence time.

This demo runs without hardware and serves as an educational example.
"""

node = QualibrationNode[Parameters, Any](
    name="06_demo_t1",
    description=description,
    parameters=Parameters(),
)


# %% Simulate experiment
@node.run_action()
def simulate_experiment(node: QualibrationNode[Parameters, Any]):
    """Simulate the T1 measurement."""
    p = node.parameters
    delays = np.linspace(0, p.max_delay, p.num_points)

    node.log(f"Simulating T1 for {p.qubits}...")
    node.log(f"Sweeping {p.num_points} delay points up to {p.max_delay} μs...")

    time_per_point = p.duration / p.num_points
    for i in range(p.num_points):
        time.sleep(time_per_point)
        if i % 10 == 0:
            node.log(f"Progress: {i}/{p.num_points} points")
        node.fraction_complete = (i + 1) / p.num_points

    node.results["raw_data"] = {
        q: {
            "delays": delays,
            "populations": _generate_t1_data(delays),
        }
        for q in p.qubits
    }
    node.log("Simulation complete!")


# %% Analyse data
@node.run_action()
def analyse_data(node: QualibrationNode[Parameters, Any]):
    """Fit exponential decay to extract T1."""
    fit_results = {}
    for q in node.parameters.qubits:
        raw = node.results["raw_data"][q]
        fit = _fit_t1(raw["delays"], raw["populations"])
        if fit["success"]:
            t1_us = fit["t1"] / 1e-6
            node.log(f"{q}: T1 = {t1_us:.2f} μs")
            node.log(f"{q}: Initial amplitude = {fit['amplitude']:.3f}")
        else:
            node.log(f"{q}: Fit failed - {fit['error']}")
        fit_results[q] = fit
    node.results["fit_results"] = fit_results
    node.outcomes = {
        q: "successful" if fit["success"] else "failed"
        for q, fit in fit_results.items()
    }


# %% Plot data
@node.run_action()
def plot_data(node: QualibrationNode[Parameters, Any]):
    """Plot T1 decay curve with fit."""
    qubits = node.parameters.qubits
    fig, axes = plt.subplots(1, len(qubits), figsize=(6 * len(qubits), 5))
    axes = [axes] if len(qubits) == 1 else axes

    for ax, qubit in zip(axes, qubits, strict=True):
        raw = node.results["raw_data"][qubit]
        fit_result = node.results["fit_results"][qubit]
        delays, populations = raw["delays"], raw["populations"]

        # Convert to microseconds for display
        delays_us = delays / 1e-6

        ax.scatter(delays_us, populations, alpha=0.6, s=40, label="Measured")

        if fit_result["success"]:
            delay_fine = np.linspace(delays[0], delays[-1], 200)
            delay_fine_us = delay_fine / 1e-6
            fit_curve = _exponential_decay(delay_fine, *fit_result["fit_params"])
            ax.plot(delay_fine_us, fit_curve, "r-", linewidth=2, label="Exponential Fit")

            # Mark T1 time
            t1_us = fit_result["t1"] / 1e-6
            ax.axvline(
                t1_us,
                color="green",
                linestyle="--",
                alpha=0.5,
                label=f"T₁ = {t1_us:.1f} μs",
            )

            # Show 1/e decay point
            decay_point = _exponential_decay(
                fit_result["t1"], *fit_result["fit_params"]
            )
            ax.plot(t1_us, decay_point, "go", markersize=10, alpha=0.7)

        ax.set_xlabel("Wait Time (μs)", fontsize=12)
        ax.set_ylabel("Excited State Population", fontsize=12)
        ax.set_title(f"T1 Decay: {qubit}", fontsize=14, fontweight="bold")
        ax.legend()
        ax.grid(True, alpha=0.3)

    plt.tight_layout()
    plt.show()
    node.results["figures"] = {"t1": fig}
    node.log("Plots generated successfully")


# %% Save results
@node.run_action()
def save_results(node: QualibrationNode[Parameters, Any]):
    """Save results."""
    node.save()
    node.log(f"Results saved. Outcomes: {node.outcomes}")



from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from typing import Any, Generic, cast

from pydantic import create_model

from qualibrate.core.models.execution_history import (
    ExecutionHistory,
    ExecutionHistoryItem,
)
from qualibrate.core.models.outcome import Outcome
from qualibrate.core.parameters import NodeParameters, RunnableParameters
from qualibrate.core.qualibration_graph import GraphElementTypeVar, QualibrationGraph
from qualibrate.core.qualibration_node import QualibrationNode
from qualibrate.core.utils.logger_m import logger
from qualibrate.core.utils.naming import get_full_class_path
from qualibrate.core.utils.type_protocols import MachineProtocol, TargetType

__all__ = ["QualibrationOrchestrator"]


class QualibrationOrchestrator(ABC, Generic[GraphElementTypeVar]):
    """
    Abstract base class for orchestrating the execution of nodes in a
    calibration graph.

    This class manages execution flow, runtime tracking, and node traversal
    within a calibration graph. It also provides methods to stop the execution,
    clean up resources, and serialize the orchestrator's current state.

    Args:
        **parameters: Keyword arguments for initializing the orchestrator
            parameters.
    """

    def __init__(self, **parameters: Any):
        self._graph: QualibrationGraph[GraphElementTypeVar] | None = None
        self._is_stopped: bool = False
        self.parameters_class = create_model(
            "OrchestratorParameters",
            __base__=RunnableParameters,
            **{  # type: ignore
                name: (type(value), value) for name, value in parameters.items()
            },
        )
        self._parameters = self.parameters_class()
        self.initial_targets: list[Any] | None = None
        self.targets: list[Any] | None = None
        self._execution_history: list[ExecutionHistoryItem] = []
        self._active_element: GraphElementTypeVar | None = None
        self.final_outcomes: dict[Any, Outcome] = {}

    @property
    def active_element(self) -> GraphElementTypeVar | None:
        """
        Gets the currently active node.

        Returns:
            Optional[QualibrationNode]: The node that is currently active, or
            None if no node is active.
        """
        return self._active_element

    @property
    def active_node(
        self,
    ) -> QualibrationNode[NodeParameters, MachineProtocol] | None:
        """
        Gets the currently active node.

        Returns:
            Optional[QualibrationNode]: The node that is currently active, or
            None if no node is active.
        """
        if self._active_element is None:
            return None

        return (
            self._active_element.active_node
            if isinstance(self._active_element, QualibrationGraph)
            else cast(
                QualibrationNode[NodeParameters, MachineProtocol],
                self._active_element,
            )
        )

    @property
    def active_element_name(self) -> str | None:
        """
        Gets the name of the currently active node.

        Returns:
            Optional[str]: The name of the active node, or None if no node
            is active.
        """
        return elem.name if (elem := self.active_node) else None

    @property
    def active_node_name(self) -> str | None:
        """
        Gets the name of the currently active node.

        Returns:
            Optional[str]: The name of the active node, or None if no node
            is active.
        """
        return node.name if (node := self.active_node) else None

    def stop(self) -> None:
        """
        Stops the execution of the orchestrator.

        Sets `_is_stopped` to True to indicate that execution should be halted.
        """
        logger.debug("Orchestrator. Stop")
        self._is_stopped = True

    def cleanup(self) -> None:
        """
        Cleans up the orchestrator state.

        Resets the internal graph, targets, execution history, active node,
        and final outcomes.
        """
        logger.debug("Orchestrator. Cleanup")
        self._graph = None
        self._is_stopped = False
        self.initial_targets = None
        self.targets = None
        self._execution_history = []
        self._active_element = None
        self.final_outcomes = {}

    def serialize(self) -> Mapping[str, Any]:
        """
        Serializes the orchestrator's current state into a dictionary
        representation.

        Returns:
            Mapping[str, Any]: A dictionary containing the class path and
            serialized parameters.
        """
        return {
            "__class__": get_full_class_path(self.__class__),
            "parameters": self.parameters_class.serialize(),
        }

    def get_execution_history(self) -> ExecutionHistory:
        """
        Gets the execution history of the orchestrator.

        Returns:
            ExecutionHistory: An object containing the execution history items.
        """
        return ExecutionHistory(items=self._execution_history)

    @abstractmethod
    def traverse_graph(
        self,
        graph: QualibrationGraph[GraphElementTypeVar],
        targets: Sequence[TargetType],
    ) -> None:
        """
        Abstract method for traversing a graph.

        This method should be implemented by subclasses to define the execution
        flow for the provided graph and targets.

        Args:
            graph (QualibrationGraph): The calibration graph to be traversed.
            targets (Sequence[Any]): The list of target nodes to start
            execution from.
        """
        pass



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, suppress

from fastapi import FastAPI

from qualibrate.app.api.sockets.tasks import update_snapshot_history_required

__all__ = ["app_lifespan"]


@asynccontextmanager
async def app_lifespan(app: FastAPI) -> AsyncIterator[None]:
    # Start periodic tasks and capture task references
    snapshot_history_task = await update_snapshot_history_required()
    try:
        yield
    finally:
        # Cancel periodic tasks on shutdown
        snapshot_history_task.cancel()
        with suppress(asyncio.CancelledError):
            await snapshot_history_task



from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from qualibrate.core.models.execution_type import ExecutionType
from qualibrate.core.models.outcome import Outcome
from qualibrate.core.storage.snapshot_json_handler import SnapshotJsonHandler
from qualibrate.core.utils.logger_m import logger

if TYPE_CHECKING:
    from qualang_tools.results import DataHandler

    from qualibrate.core.qualibration_graph import QualibrationGraph

__all__ = ["WorkflowSnapshotManager"]


class WorkflowSnapshotManager:
    """Manages workflow snapshot lifecycle independent of orchestrator.

    This class handles the creation, tracking, and finalization of
    workflow snapshots, decoupling storage concerns from the orchestrator.

    Args:
        root_data_folder: The root folder where data should be saved.
        data_handler: DataHandler instance for creating snapshots.
    """

    def __init__(
        self,
        root_data_folder: Path,
        data_handler: "DataHandler",
    ):
        self.root_data_folder = root_data_folder
        self.data_handler = data_handler
        self.json_handler = SnapshotJsonHandler(root_data_folder)

    def _clean_data_handler(self) -> None:
        """Reset data handler path state."""
        self.data_handler.path = None
        self.data_handler.path_properties = None

    def start_workflow_snapshot(
        self,
        graph: "QualibrationGraph[Any]",
        workflow_parent_id: int | None = None,
    ) -> int:
        """Create a workflow snapshot at the start of graph execution.

        Args:
            graph: The QualibrationGraph being executed.
            workflow_parent_id: ID of parent workflow if this is a nested graph.

        Returns:
            The snapshot ID of the created workflow snapshot.
        """
        logger.info(f"Saving workflow snapshot start for {graph.name}")

        self.data_handler.name = graph.name
        self._clean_data_handler()

        run_start = datetime.now().astimezone()

        # Prepare workflow node data
        self.data_handler.node_data = {
            "parameters": {
                "model": (graph.full_parameters.model_dump(mode="json") if graph.full_parameters else {}),
            },
            "outcomes": {},
        }

        metadata: dict[str, Any] = {
            "description": graph.description,
            "run_start": run_start.isoformat(timespec="milliseconds"),
            "type_of_execution": ExecutionType.workflow.value,
            "children": [],  # Will be populated as child nodes complete
            "status": "running",
        }
        if workflow_parent_id is not None:
            metadata["workflow_parent_id"] = workflow_parent_id

        node_contents = self.data_handler.generate_node_contents(metadata=metadata)
        self.data_handler.save_data(
            data={},
            name=graph.name,
            node_contents=node_contents,
        )
        snapshot_idx = cast(int, node_contents["id"])
        logger.info(f"Created workflow snapshot {snapshot_idx} for {graph.name}")
        return snapshot_idx

    def finalize_workflow_snapshot(
        self,
        graph: "QualibrationGraph[Any]",
        workflow_snapshot_idx: int,
        children_ids: list[int],
        outcomes: dict[str, Any],
        status: str = "finished",
    ) -> bool:
        """Update a workflow snapshot at the end of graph execution.

        Args:
            graph: The QualibrationGraph that finished executing.
            workflow_snapshot_idx: The snapshot ID of the workflow to update.
            children_ids: List of child snapshot IDs.
            outcomes: Aggregated outcomes from the workflow execution.
            status: Final status of the workflow ("finished" or "error").

        Returns:
            True if update succeeded, False otherwise.
        """
        logger.info(f"Finalizing workflow snapshot {workflow_snapshot_idx} for {graph.name}")

        node_json_path = self.json_handler.get_node_json_path(workflow_snapshot_idx)
        if node_json_path is None:
            return False

        content = self.json_handler.read_node_json(node_json_path)
        if content is None:
            return False

        run_end = datetime.now().astimezone()

        # Update metadata
        if "metadata" not in content:
            content["metadata"] = {}
        content["metadata"]["run_end"] = run_end.isoformat(timespec="milliseconds")
        content["metadata"]["children"] = children_ids
        content["metadata"]["status"] = status

        # Update outcomes in data
        if "data" not in content:
            content["data"] = {}
        content["data"]["outcomes"] = {k: v.value if isinstance(v, Outcome) else v for k, v in outcomes.items()}

        success = self.json_handler.write_node_json(node_json_path, content)
        if success:
            logger.info(
                f"Finalized workflow snapshot {workflow_snapshot_idx} for graph {graph.name} with status {status}"
            )
        return success

    def track_child_snapshot(
        self,
        workflow_snapshot_idx: int,
        child_id: int,
    ) -> bool:
        """Append a child snapshot ID to a workflow's children list.

        Args:
            workflow_snapshot_idx: The workflow snapshot ID to update.
            child_id: The child snapshot ID to add.

        Returns:
            True if update succeeded, False otherwise.
        """
        node_json_path = self.json_handler.get_node_json_path(workflow_snapshot_idx)
        if node_json_path is None:
            return False

        content = self.json_handler.read_node_json(node_json_path)
        if content is None:
            return False

        # Ensure metadata and children exist
        if "metadata" not in content:
            content["metadata"] = {}
        if "children" not in content["metadata"]:
            content["metadata"]["children"] = []

        # Append child if not already present
        if child_id not in content["metadata"]["children"]:
            content["metadata"]["children"].append(child_id)

        return self.json_handler.write_node_json(node_json_path, content)

    def set_snapshot_workflow_parent(
        self,
        snapshot_idx: int,
        workflow_parent_id: int,
    ) -> bool:
        """Set the workflow_parent_id for a snapshot.

        Args:
            snapshot_idx: The snapshot ID to update.
            workflow_parent_id: The parent workflow's snapshot ID.

        Returns:
            True if update succeeded, False otherwise.
        """
        node_json_path = self.json_handler.get_node_json_path(snapshot_idx)
        if node_json_path is None:
            return False

        content = self.json_handler.read_node_json(node_json_path)
        if content is None:
            return False

        # Ensure metadata exists and set workflow_parent_id
        if "metadata" not in content:
            content["metadata"] = {}
        content["metadata"]["workflow_parent_id"] = workflow_parent_id
        # Only set type_of_execution to node if not already set
        # (preserves workflow type for nested subgraphs)
        if "type_of_execution" not in content["metadata"]:
            content["metadata"]["type_of_execution"] = ExecutionType.node.value

        return self.json_handler.write_node_json(node_json_path, content)



from qualibrate.core.parameters import (
    ExecutionParameters,
    GraphParameters,
    NodeParameters,
    NodesParameters,
)
from qualibrate.core.qualibration_graph import QualibrationGraph
from qualibrate.core.qualibration_library import QualibrationLibrary
from qualibrate.core.qualibration_node import QualibrationNode
from qualibrate.core.utils import logger

__all__ = [
    "ExecutionParameters",
    "GraphParameters",
    "logger",
    "NodeParameters",
    "NodesParameters",
    "QualibrationLibrary",
    "QualibrationNode",
    "QualibrationGraph",
]



import json
import logging
from collections import deque
from collections.abc import Callable
from datetime import datetime
from typing import Any

from qualibrate.core.utils.logger_utils.filters import filter_log_date

__all__ = ["InMemoryLogHandler"]


class InMemoryLogHandler(logging.Handler):
    default_asctime_log_format = "%Y-%m-%d %H:%M:%S,%f"

    def __init__(self, max_logs: int = 10_000):
        super().__init__()
        self.logs: deque[dict[str, Any]] = deque(maxlen=max_logs)  # Automatically rotates
        self._broadcast_callback: Callable[[dict[str, Any]], None] | None = None

    def set_broadcast_callback(self, callback: Callable[[dict[str, Any]], None] | None) -> None:
        """Set a callback function to be called when a log entry is emitted.

        The callback receives the parsed log entry dict and can be used
        to broadcast logs via WebSocket in real-time.
        """
        self._broadcast_callback = callback

    def emit(self, record: logging.LogRecord) -> None:
        try:
            msg = self.format(record)
            parsed = json.loads(msg)
            if asctime := parsed.get("asctime"):
                parsed["asctime"] = datetime.strptime(asctime, self.__class__.default_asctime_log_format)
            self.logs.append(parsed)
            if self._broadcast_callback is not None:
                self._broadcast_callback(parsed)
        except Exception:
            self.handleError(record)

    def get_logs(
        self,
        after: datetime | None = None,
        before: datetime | None = None,
        num_entries: int = 100,
    ) -> list[dict[str, Any]]:
        filtered_logs = [log for log in reversed(self.logs) if filter_log_date(log)]
        return list(reversed(filtered_logs[:num_entries]))



from .snapshot_json_handler import SnapshotJsonHandler
from .storage_manager import StorageManager
from .workflow_snapshot_manager import WorkflowSnapshotManager

__all__ = ["SnapshotJsonHandler", "StorageManager", "WorkflowSnapshotManager"]



import operator
import sys
from functools import reduce
from typing import Annotated

from fastapi import HTTPException, Query

from qualibrate.app.api.core.domain.bases.snapshot import SnapshotLoadTypeFlag

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


__all__ = ["SnapshotLoadTypeStr", "parse_load_type_flag"]


class SnapshotLoadTypeStr(StrEnum):
    Minified = "Minified"
    Metadata = "Metadata"
    DataWithoutRefs = "DataWithoutRefs"
    DataWithMachine = "DataWithMachine"
    DataWithResults = "DataWithResults"
    DataWithResultsWithImgs = "DataWithResultsWithImgs"
    Full = "Full"


def _parse_snapshot_load_type(
    name: SnapshotLoadTypeStr,
) -> SnapshotLoadTypeFlag:
    try:
        return SnapshotLoadTypeFlag[name]
    except KeyError:
        raise HTTPException(status_code=400, detail=f"Invalid load type '{name}'") from None


def parse_load_type_flag(
    load_type_flag: Annotated[
        list[SnapshotLoadTypeStr] | None,
        Query(description="Bitwise flags specifying which snapshot parts to load."),
    ] = None,
) -> SnapshotLoadTypeFlag:
    if load_type_flag is None:
        return SnapshotLoadTypeFlag.Minified
    load_types = [_parse_snapshot_load_type(t) for t in load_type_flag]
    return reduce(
        operator.or_,
        load_types,
        SnapshotLoadTypeFlag.Minified,
    )



from abc import ABC, abstractmethod
from collections.abc import Sequence
from pathlib import Path

from qualibrate_config.core.project.create import create_project
from qualibrate_config.models import QualibrateConfig

from qualibrate.app.api.core.domain.bases.base_with_settings import (
    DomainWithConfigBase,
)
from qualibrate.app.api.core.models.project import Project
from qualibrate.app.api.exceptions.classes.values import QValueException


class ProjectsManagerBase(DomainWithConfigBase, ABC):
    def __init__(self, settings: QualibrateConfig, config_path: Path):
        super().__init__(settings)
        self._config_path = config_path

    @property
    def project(self) -> str | None:
        return self._active_project_getter()

    @project.setter
    def project(self, value: str) -> None:
        self._active_project_setter(value)

    def _active_project_getter(self) -> str | None:
        return self._settings.project

    @abstractmethod
    def _active_project_setter(self, value: str) -> None:
        pass

    def create(
        self,
        project_name: str,
        storage_location: Path | None = None,
        calibration_library_folder: Path | None = None,
        quam_state_path: Path | None = None,
    ) -> str:
        try:
            create_project(
                self._config_path,
                project_name,
                storage_location,
                calibration_library_folder,
                quam_state_path,
            )
        except ValueError as e:
            raise QValueException(f"Failed to create project '{project_name}'") from e
        return project_name

    @abstractmethod
    def list(self) -> Sequence[Project]:
        pass



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from types import ModuleType

__all__ = [
    "get_module_name",
    "import_from_path",
    "import_from_path_exec",
    "import_from_path_importlib",
]


def get_module_name(file_path: Path) -> str:
    """Create module name from file path."""
    return f"_node_{file_path.stem}"


def import_from_path(module_name: str, file_path: Path) -> ModuleType:
    try:
        return import_from_path_importlib(module_name, file_path)
    except UnicodeError:
        return import_from_path_exec(module_name, file_path)


def import_from_path_importlib(module_name: str, file_path: Path) -> ModuleType:
    """Import a module given its name and file path."""
    spec = spec_from_file_location(module_name, file_path)
    if spec is None:
        raise RuntimeError(f"Can't read spec from {file_path}")
    module = module_from_spec(spec)
    if spec.loader is None:
        raise RuntimeError(f"Can't get loader from spec of file {file_path}")
    spec.loader.exec_module(module)
    return module


def import_from_path_exec(module_name: str, file_path: Path, encoding: str = "utf-8") -> ModuleType:
    """Import a module from a file with enforced UTF-8 encoding."""
    source = file_path.read_text(encoding=encoding)
    module = ModuleType(module_name)
    module.__file__ = str(file_path)
    exec(compile(source, str(file_path), "exec"), module.__dict__)
    return module



import os
from functools import lru_cache
from itertools import product
from pathlib import Path
from typing import Annotated

from fastapi import Depends
from qualibrate_config.models import QualibrateCompositeConfig, QualibrateConfig
from qualibrate_config.resolvers import (
    get_qualibrate_config,
    get_qualibrate_config_path,
)

from qualibrate.composite.config.vars import (
    CONFIG_PATH_ENV_NAME,
    CORS_ORIGINS_ENV_NAME,
    ROOT_PATH_ENV_NAME,
)

__all__ = [
    "get_config_path",
    "get_cors_origin",
    "get_settings",
    "get_root_path",
    "get_qualibrate_config",
    "get_composite_settings",
]


@lru_cache
def get_config_path() -> Path:
    path = os.environ.get(CONFIG_PATH_ENV_NAME)
    if path is not None:
        return Path(path)
    return get_qualibrate_config_path()


@lru_cache
def get_cors_origin() -> list[str]:
    cors_env = os.environ.get(CORS_ORIGINS_ENV_NAME)
    if cors_env is not None:
        return list(cors_env.split(","))
    return list(f"http://{host}:{port}" for host, port in product(["localhost", "127.0.0.1"], [1234, 8000, 8001]))


@lru_cache
def get_root_path() -> str:
    return os.environ.get(ROOT_PATH_ENV_NAME, "")


@lru_cache
def get_settings(
    config_path: Annotated[Path, Depends(get_config_path)],
) -> QualibrateConfig:
    return get_qualibrate_config(config_path)


def get_composite_settings(
    config: Annotated[QualibrateConfig, Depends(get_settings)],
) -> QualibrateCompositeConfig:
    if config.composite is None:
        raise RuntimeError("Composite part not specified in config.toml")
    return config.composite



from collections.abc import Sequence
from typing import Any

from qualibrate.app.api.core.domain.bases.branch import BranchBase
from qualibrate.app.api.core.domain.bases.node import NodeBase
from qualibrate.app.api.core.domain.bases.root import RootBase
from qualibrate.app.api.core.domain.bases.snapshot import SnapshotBase
from qualibrate.app.api.core.domain.local_storage.branch import (
    BranchLocalStorage,
)
from qualibrate.app.api.core.domain.local_storage.node import NodeLocalStorage
from qualibrate.app.api.core.domain.local_storage.snapshot import (
    SnapshotLocalStorage,
)
from qualibrate.app.api.core.domain.local_storage.utils.node_utils import (
    find_nodes_ids_by_filter,
)
from qualibrate.app.api.core.models.snapshot import SnapshotSearchResult
from qualibrate.app.api.core.types import (
    IdType,
    PageFilter,
    SearchWithIdFilter,
)
from qualibrate.app.api.exceptions.classes.storage import QFileNotFoundException

__all__ = ["RootLocalStorage"]


class RootLocalStorage(RootBase):
    def get_branch(self, branch_name: str) -> BranchBase:
        return BranchLocalStorage(branch_name, settings=self._settings)

    def _get_latest_node_id(self, error_msg: str) -> IdType:
        id = next(
            find_nodes_ids_by_filter(
                self._settings.storage.location,
                project_name=self._settings.project,
                descending=True,
            ),
            None,
        )
        if id is None:
            raise QFileNotFoundException(f"There is no {error_msg}")
        return id

    def get_snapshot(self, id: IdType | None = None) -> SnapshotBase:
        if id is None:
            id = self._get_latest_node_id("snapshot")
        return SnapshotLocalStorage(id, settings=self._settings)

    def get_node(self, id: IdType | None = None) -> NodeBase:
        if id is None:
            id = self._get_latest_node_id("node")
        return NodeLocalStorage(id, settings=self._settings)

    def get_latest_snapshots(
        self,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        descending: bool = False,
        include_outcomes: bool = False,
    ) -> tuple[int, Sequence[SnapshotBase]]:
        return BranchLocalStorage("main", settings=self._settings).get_latest_snapshots(
            pages_filter=pages_filter,
            search_filter=search_filter,
            descending=descending,
            include_outcomes=include_outcomes,
        )

    def get_latest_nodes(
        self,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        descending: bool = False,
    ) -> tuple[int, Sequence[NodeBase]]:
        return BranchLocalStorage("main", settings=self._settings).get_latest_nodes(
            pages_filter=pages_filter,
            search_filter=search_filter,
            descending=descending,
        )

    def search_snapshot(
        self,
        search_filter: SearchWithIdFilter,
        data_path: Sequence[str | int],
        descending: bool = False,
    ) -> Any:
        _, data = self.search_snapshots_data(
            search_filter=search_filter,
            pages_filter=PageFilter(page=1, per_page=10000),
            data_path=data_path,
            descending=descending,
            filter_no_change=False,
        )
        return data

    def search_snapshots_data(
        self,
        *,
        pages_filter: PageFilter,
        search_filter: SearchWithIdFilter | None = None,
        data_path: Sequence[str | int],
        filter_no_change: bool = True,
        descending: bool = False,
    ) -> tuple[int, Sequence[SnapshotSearchResult]]:
        return BranchLocalStorage("main", settings=self._settings).search_snapshots_data(
            search_filter=search_filter,
            pages_filter=pages_filter,
            data_path=data_path,
            descending=descending,
            filter_no_change=filter_no_change,
        )



import logging
from datetime import date
from typing import Annotated

from fastapi import Depends, Query

from qualibrate.app.api.core.domain.bases.snapshot import (
    SnapshotLoadType,
    SnapshotLoadTypeFlag,
    SnapshotLoadTypeToLoadTypeFlag,
)
from qualibrate.app.api.core.types import (
    IdType,
    PageFilter,
    SearchFilter,
    SearchWithIdFilter,
)
from qualibrate.app.api.routes.utils.snapshot_load_type import (
    parse_load_type_flag,
)

logger = logging.getLogger(__name__)


def get_page_filter(
    page: int = Query(
        1,
        gt=0,
        description="Page number (1-based).",
    ),
    per_page: int = Query(
        50,
        gt=0,
        description="Number of items per page.",
    ),
) -> PageFilter:
    return PageFilter(page=page, per_page=per_page)


def get_search_filter(
    snapshot_name: Annotated[
        str | None,
        Query(
            alias="name",
            description="Exact snapshot name to match.",
        ),
    ] = None,
    name_part: Annotated[
        str | None,
        Query(description="Substring to match within snapshot name."),
    ] = None,
    min_node_id: Annotated[
        IdType,
        Query(description="Lower bound (inclusive) for node ID range."),
    ] = 1,
    max_node_id: Annotated[
        IdType | None,
        Query(description="Upper bound (inclusive) for node ID range."),
    ] = None,
    min_date: Annotated[
        date | None,
        Query(description="Earliest snapshot date (inclusive)."),
    ] = None,
    max_date: Annotated[
        date | None,
        Query(description="Latest snapshot date (inclusive)."),
    ] = None,
    tag_name: Annotated[
        list[str] | None,
        Query(
            description=(
                "Filter by tag names. Snapshots must have ALL specified tags "
                "(AND logic). Can specify multiple: ?tag_name=t1&tag_name=t2"
            )
        ),
    ] = None,
) -> SearchFilter:
    search_filter = SearchFilter(
        name=snapshot_name,
        name_part=name_part,
        min_node_id=min_node_id,
        max_node_id=max_node_id,
        min_date=min_date,
        max_date=max_date,
        tags=tag_name,
    )
    logger.debug(
        f"Created SearchFilter: name={snapshot_name!r}, name_part={name_part!r}, "
        f"min_node_id={min_node_id}, max_node_id={max_node_id}, "
        f"min_date={min_date}, max_date={max_date}, tags={tag_name}"
    )
    return search_filter


def get_search_with_id_filter(
    search_filter: Annotated[SearchFilter, Depends(get_search_filter)],
    id: Annotated[
        IdType | None,
        Query(description=("Exact snapshot ID to match (overrides name/date filters if set).")),
    ] = None,
) -> SearchWithIdFilter:
    return SearchWithIdFilter(
        id=id,
        **search_filter.model_dump(),
    )


def get_snapshot_load_type_flag(
    load_type: Annotated[SnapshotLoadType | None, Query(deprecated="use load_type_flag")] = None,
    load_type_flag: Annotated[SnapshotLoadTypeFlag, Depends(parse_load_type_flag)] = SnapshotLoadTypeFlag.Metadata,
) -> SnapshotLoadTypeFlag:
    return load_type_flag if load_type is None else SnapshotLoadTypeToLoadTypeFlag[load_type]



"""
11 - Tune-up with retries

This example demonstrates:
- `loop`: How to automatically retry a calibration node.
- `connect_on_failure`: How to define a fallback path for error handling.
"""
from qualibrate.core import GraphParameters, QualibrationGraph, QualibrationLibrary

# Get the active library of pre-defined calibration nodes
library = QualibrationLibrary.get_active_library()


class TuneupParameters(GraphParameters):
    """Parameters for the tune-up graph, specifying the qubits to be used."""
    qubits: list[str] = ["q1"]


# Use the context manager to build the graph
with QualibrationGraph.build(
    "11_tuneup_with_retries",
    parameters=TuneupParameters(),
    # description="A graph that retries a Rabi experiment and has a failure path.",
) as graph:
    # 1. Initial Rabi experiment
    # We will try to run a Rabi experiment. If it fails, we want to retry.
    rabi_node = library.nodes["02_demo_rabi"]
    rabi_node.name = "rabi_with_retries"
    graph.add_node(rabi_node)

    # Use loop to retry the node up to 3 times.
    graph.loop(rabi_node, max_iterations=3)

    # 2. Refined Rabi experiment (success path)
    # This node will only run if the initial Rabi experiment succeeds.
    refined_rabi_node = library.nodes["04_demo_rabi_refined"]
    graph.add_node(refined_rabi_node)
    graph.connect(rabi_node, refined_rabi_node)

    # 3. Ramsey experiment (success path)
    # This node will run after the refined Rabi, continuing the successful path.
    ramsey_node = library.nodes["05_demo_ramsey"]
    graph.add_node(ramsey_node)
    graph.connect(refined_rabi_node, ramsey_node)

    # 4. Failure handling node
    # This is a placeholder for a failure-handling procedure. It could be a
    # simple wait, a notification, or another calibration to reset the qubit.
    # For this example, we use a simple T1 measurement as a placeholder.
    failure_handler_node = library.nodes["06_demo_t1"]
    failure_handler_node.name = "failure_handler"
    graph.add_node(failure_handler_node)

    # Use connect_on_failure to specify that if 'rabi_with_retries' fails
    # after all retries, the graph should proceed to the failure handler.
    graph.connect_on_failure(rabi_node, failure_handler_node)

# The graph is now finalized.
if __name__ == "__main__":
    # When executed, if `rabi_with_retries` fails, it will be retried 3 times.
    # If it still fails, the execution will follow the edge to `failure_handler`.
    # If it succeeds, execution proceeds to `refined_rabi_node` and then `ramsey_node`.
    print(f"Running graph: {graph.name}...")
    result = graph.run()
    print("Graph execution finished.")
    print(result)



from typing import Annotated

from fastapi import APIRouter, Depends
from starlette.websockets import WebSocket, WebSocketDisconnect

from qualibrate.runner.core.app.ws_manager import (
    SocketConnectionManagerList,
    SocketConnectionManagerMapping,
)
from qualibrate.runner.core.app.ws_managers import (
    get_execution_history_socket_manager,
    get_output_logs_socket_manager,
    get_run_status_socket_manager,
)

common_ws_router = APIRouter()


@common_ws_router.websocket("/run_status")
async def run_status_subscribe(
    websocket: WebSocket,
    *,
    manager: Annotated[SocketConnectionManagerList, Depends(get_run_status_socket_manager)],
) -> None:
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        manager.disconnect(websocket)


@common_ws_router.websocket("/workflow_execution_history")
async def workflow_execution_history_subscribe(
    websocket: WebSocket,
    *,
    reverse: bool = True,
    manager: Annotated[
        SocketConnectionManagerMapping[bool],
        Depends(get_execution_history_socket_manager),
    ],
) -> None:
    await manager.connect(reverse, websocket)
    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        manager.disconnect(reverse, websocket)


@common_ws_router.websocket("/output_logs")
async def output_logs_subscribe(
    websocket: WebSocket,
    *,
    manager: Annotated[SocketConnectionManagerList, Depends(get_output_logs_socket_manager)],
) -> None:
    """WebSocket endpoint for real-time log streaming.

    Logs are pushed to subscribers as they are generated.
    """
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        manager.disconnect(websocket)



from qualibrate import QualibrationGraph, QualibrationNode
from qualibrate.core.models.execution_history import ExecutionHistory
from qualibrate.runner.config import State
from qualibrate.runner.core.models.active_run import (
    RunStatus,
    RunStatusGraph,
    RunStatusNode,
)
from qualibrate.runner.core.models.enums import RunStatusEnum
from qualibrate.runner.core.models.last_run import LastRun
from qualibrate.runner.core.models.run_results import RunResults
from qualibrate.runner.core.types import QGraphType, QNodeType

__all__ = [
    "get_graph_and_node_run_status",
    "get_graph_execution_history",
    "get_node_run_status",
    "get_run_status",
]


def get_node_status_enum(
    last_run: LastRun,
    node: QNodeType,
    graph: QGraphType | None,
) -> RunStatusEnum:
    if graph is None:
        return last_run.status
    if node.run_summary is None:
        return RunStatusEnum.RUNNING
    if node.run_summary.error is None:
        return RunStatusEnum.FINISHED
    return RunStatusEnum.ERROR


def get_node_run_status(last_run: LastRun, node: QNodeType, graph: QGraphType | None) -> RunStatusNode:
    return RunStatusNode(
        name=node.name,
        description=node.description,
        parameters=node.parameters.model_dump(mode="json"),
        id=node.snapshot_idx or last_run.idx,
        status=get_node_status_enum(last_run, node, graph),
        run_start=node.run_start,
        current_action=node.action_label,
        run_end=last_run.completed_at,
        percentage_complete=node.fraction_complete * 100,
        run_results=(RunResults.model_validate(node.run_summary.model_dump()) if node.run_summary else None),
    )


def get_graph_and_node_run_status(
    last_run: LastRun, graph: QGraphType, node: QNodeType | None
) -> tuple[RunStatusGraph, RunStatusNode | None]:
    graph_status = RunStatusGraph(
        name=graph.name,
        description=graph.description,
        parameters=graph.full_parameters.model_dump(mode="json"),
        status=last_run.status,
        run_start=graph.run_start,
        run_end=last_run.completed_at,
        finished_nodes=graph.completed_count(),
        total_nodes=len(graph._nodes),
        run_results=(RunResults.model_validate(graph.run_summary.model_dump()) if graph.run_summary else None),
    )
    orchestrator = graph._orchestrator
    execution_history = orchestrator.get_execution_history().items if orchestrator else []
    if node is not None or orchestrator is None or len(execution_history) == 0:
        return graph_status, None
    node_hist = execution_history[-1]
    node_status = RunStatusNode(
        name=node_hist.metadata.name,
        description=node_hist.metadata.description,
        parameters=node_hist.data.parameters.model_dump(mode="json"),
        id=node_hist.id,
        status=RunStatusEnum.FINISHED,
        run_start=node_hist.metadata.run_start,
        run_end=node_hist.metadata.run_end,
        percentage_complete=100,
    )
    return graph_status, node_status


def get_run_status(state: State) -> RunStatus:
    last_run = state.last_run
    if last_run is None:
        return RunStatus()
    if isinstance(state.run_item, QualibrationNode):
        node: QNodeType | None = state.run_item
        graph: QGraphType | None = None
    elif isinstance(state.run_item, QualibrationGraph):
        graph = state.run_item
        node = graph.active_node
    else:
        node = None
        graph = None

    node_status: RunStatusNode | None = None
    graph_status: RunStatusGraph | None = None
    if node:
        node_status = get_node_run_status(last_run, node, graph)
    if graph:
        graph_status, node_graph_status = get_graph_and_node_run_status(last_run, graph, node)
        node_status = node_status or node_graph_status
    return RunStatus(
        is_running=state.is_running,
        runnable_type=last_run.runnable_type,
        node=node_status,
        graph=graph_status,
    )


def get_graph_execution_history(state: State, reverse: bool = False) -> ExecutionHistory | None:
    if not isinstance(state.run_item, QualibrationGraph):
        return None
    graph: QGraphType = state.run_item
    orch = graph._orchestrator
    if orch is None:
        raise RuntimeError("No graph orchestrator")
    history: ExecutionHistory = orch.get_execution_history()
    if reverse:
        history.items = list(reversed(history.items))
    return history



import logging
from pathlib import Path

from qualibrate.core.parameters import NodeParameters, RunnableParameters
from qualibrate.core.q_runnnable import QRunnable
from qualibrate.core.qualibration_node import QualibrationNode
from qualibrate.core.utils.type_protocols import MachineProtocol

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    from qualibrate.core.qualibration_library import QualibrationLibrary

    library_folder = Path(__file__).parent.parent / "calibrations"
    library = QualibrationLibrary[
        QualibrationNode[NodeParameters, MachineProtocol],
        QRunnable[RunnableParameters, RunnableParameters],
    ](library_folder=library_folder)

    print(library.nodes)
    print(library.graphs)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any


class BaseLoader(ABC):
    """
    Abstract base class for file loaders.

    Attributes:
        file_extensions: A tuple of file extensions supported by the loader.
    """

    file_extensions: tuple[str, ...]

    @classmethod
    def is_loader_support_extension(cls, extension: str) -> bool:
        """
        Checks if the loader supports a specific file extension.

        Args:
            extension: The file extension to check.

        Returns:
            True if the extension is supported, False otherwise.
        """
        return extension.lower() in cls.file_extensions

    @classmethod
    def validate_file_exists(cls, file_path: Path) -> None:
        if not file_path.is_file():
            raise FileNotFoundError(f"Can't load file '{file_path}' by loader {cls.__name__}")

    @abstractmethod
    def load(self, path: Path, **kwargs: Any) -> Any:
        """
        Abstract method to load a file.

        Args:
            path: The path to the file to load.
            **kwargs: Additional arguments for file loading.

        Returns:
            The loaded content.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.
        """
        raise NotImplementedError



from .models import State
from .resolvers import get_cl_settings, get_config_path
from .vars import (
    CONFIG_KEY,
    CONFIG_PATH_ENV_NAME,
    DEFAULT_QUALIBRATE_RUNNER_CONFIG_FILENAME,
)

# CONFIG KEY seems deprecated
__all__ = [
    "CONFIG_KEY",
    "DEFAULT_QUALIBRATE_RUNNER_CONFIG_FILENAME",
    "CONFIG_PATH_ENV_NAME",
    "State",
    "get_config_path",
    "get_cl_settings",
]



"""These models based on RunSummary of qualibrate-core but exclude duplications
of RunStatus fields"""

from collections.abc import Mapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field

from qualibrate.core.models.outcome import Outcome
from qualibrate.core.utils.type_protocols import TargetType
from qualibrate.runner.core.models.common import RunError


class RunResults(BaseModel):
    model_config = ConfigDict(extra="ignore")

    parameters: dict[str, Any]
    outcomes: dict[TargetType, Outcome]
    error: RunError | None = None

    initial_targets: list[TargetType] = Field(default_factory=list)
    successful_targets: list[TargetType] = Field(default_factory=list)
    failed_targets: list[TargetType] = Field(default_factory=list)
    state_updates: Mapping[str, Any] = Field(default_factory=dict)



from qualibrate.core.models.run_summary.base import BaseRunSummary
from qualibrate.core.parameters import NodeParameters

__all__ = ["NodeRunSummary"]


class NodeRunSummary(BaseRunSummary):
    snapshot_idx: int | None = None
    parameters: NodeParameters



from functools import cache

from qualibrate.runner.core.app.ws_manager import (
    SocketConnectionManagerList,
    SocketConnectionManagerMapping,
)


@cache
def get_run_status_socket_manager() -> SocketConnectionManagerList:
    return SocketConnectionManagerList()


@cache
def get_execution_history_socket_manager() -> SocketConnectionManagerMapping[bool]:
    return SocketConnectionManagerMapping[bool]()


@cache
def get_output_logs_socket_manager() -> SocketConnectionManagerList:
    return SocketConnectionManagerList()



"""Migration script to update existing snapshots with type_of_execution field.

This script scans all node.json files in the data directory and adds:
- type_of_execution: "node" (default for all existing snapshots)

For new snapshots created with workflows, the orchestrator will set
type_of_execution to "workflow" and populate the children list.

Usage:
    python -m qualibrate.app.cli.migrate_snapshots [--dry-run] [--data-path PATH]
"""

import argparse
import json
import sys
from pathlib import Path

from qualibrate_config.models import StorageType

from qualibrate.core.models.execution_type import ExecutionType


def migrate_snapshot(node_json_path: Path, dry_run: bool = False) -> bool:
    """Migrate a single snapshot's node.json file.

    Args:
        node_json_path: Path to the node.json file.
        dry_run: If True, don't write changes, just report what would change.

    Returns:
        True if the file was (or would be) modified, False otherwise.
    """
    try:
        with node_json_path.open("r") as f:
            content = json.load(f)
    except (json.JSONDecodeError, OSError) as e:
        print(f"  ERROR: Failed to read {node_json_path}: {e}")
        return False

    modified = False

    # Ensure metadata exists
    if "metadata" not in content:
        content["metadata"] = {}
        modified = True

    # Add type_of_execution if not present
    if "type_of_execution" not in content["metadata"]:
        content["metadata"]["type_of_execution"] = ExecutionType.node.value
        modified = True

    if modified and not dry_run:
        try:
            with node_json_path.open("w") as f:
                json.dump(content, f, indent=2, default=str)
        except OSError as e:
            print(f"  ERROR: Failed to write {node_json_path}: {e}")
            return False

    return modified


def find_node_json_files(data_path: Path) -> list[Path]:
    """Find all node.json files in the data directory.

    Args:
        data_path: Root data directory.

    Returns:
        List of paths to node.json files.
    """
    return list(data_path.glob("*/*/node.json"))


def run_migration(
    data_path: Path,
    dry_run: bool = False,
    verbose: bool = False,
) -> tuple[int, int, int]:
    """Run the migration on all snapshots.

    Args:
        data_path: Root data directory.
        dry_run: If True, don't write changes.
        verbose: If True, print details for each file.

    Returns:
        Tuple of (total_files, modified_files, error_count).
    """
    node_json_files = find_node_json_files(data_path)
    total = len(node_json_files)
    modified = 0
    errors = 0

    print(f"Found {total} snapshot files to process")
    if dry_run:
        print("DRY RUN - no changes will be written")
    print()

    for i, node_json_path in enumerate(node_json_files, 1):
        if verbose:
            print(f"[{i}/{total}] Processing {node_json_path.relative_to(data_path)}")

        try:
            was_modified = migrate_snapshot(node_json_path, dry_run)
            if was_modified:
                modified += 1
                if verbose:
                    action = "Would modify" if dry_run else "Modified"
                    print(f"  {action}")
        except Exception as e:
            errors += 1
            print(f"  ERROR: {e}")

    return total, modified, errors


def main() -> int:
    """Main entry point for the migration script."""
    parser = argparse.ArgumentParser(description="Migrate existing snapshots to add type_of_execution field.")
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Don't write changes, just report what would be modified.",
    )
    parser.add_argument(
        "--data-path",
        type=Path,
        help="Path to the data directory. If not provided, uses config.",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="Print details for each file.")
    args = parser.parse_args()

    # Get data path from config if not provided
    data_path = args.data_path
    if data_path is None:
        try:
            from qualibrate_config.resolvers import (
                get_qualibrate_config,
                get_qualibrate_config_path,
            )

            q_config_path = get_qualibrate_config_path()
            qs = get_qualibrate_config(q_config_path)

            if qs.storage.type != StorageType.local_storage:
                print(f"ERROR: Migration only supports local_storage, got {qs.storage.type}")
                return 1

            data_path = qs.storage.location
        except Exception as e:
            print(f"ERROR: Failed to get config: {e}")
            print("Please provide --data-path explicitly.")
            return 1

    if not data_path.is_dir():
        print(f"ERROR: Data path does not exist: {data_path}")
        return 1

    print(f"Data path: {data_path}")
    print()

    total, modified, errors = run_migration(data_path, dry_run=args.dry_run, verbose=args.verbose)

    print()
    print("=" * 50)
    print(f"Total files:    {total}")
    print(f"Modified:       {modified}")
    print(f"Errors:         {errors}")
    if args.dry_run:
        print()
        print("This was a dry run. Run without --dry-run to apply changes.")

    return 0 if errors == 0 else 1


if __name__ == "__main__":
    sys.exit(main())



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager, suppress
from datetime import datetime
from typing import Any

from fastapi import FastAPI

from qualibrate.core.utils.logger_m import logger
from qualibrate.runner.api.sockets.tasks import execution_history, run_status
from qualibrate.runner.core.app.ws_managers import get_output_logs_socket_manager

__all__ = ["app_lifespan"]


def _setup_log_broadcasting(loop: asyncio.AbstractEventLoop) -> None:
    """Wire up the logger to broadcast logs via WebSocket.

    Args:
        loop: The event loop to use for scheduling broadcasts.
              Must be passed because node.run() executes synchronously
              and asyncio.get_running_loop() won't work from sync context.
    """
    manager = get_output_logs_socket_manager()
    handler = logger.in_memory_handler

    def broadcast_log(log_entry: dict[str, Any]) -> None:
        # Schedule the async broadcast from the sync emit() call
        # Use run_coroutine_threadsafe since we might be called from sync context
        # while the event loop is blocked (e.g., during node.run())
        try:
            # Convert datetime to ISO string for JSON serialization
            json_safe_entry = log_entry.copy()
            if isinstance(json_safe_entry.get("asctime"), datetime):
                json_safe_entry["asctime"] = json_safe_entry["asctime"].isoformat()
            num_connections = len(manager.active_connections)
            if num_connections > 0:
                # Use run_coroutine_threadsafe to schedule on the saved event loop
                asyncio.run_coroutine_threadsafe(manager.broadcast(json_safe_entry), loop)
        except Exception:
            # Log errors but don't crash the logging system
            pass

    handler.set_broadcast_callback(broadcast_log)


@asynccontextmanager
async def app_lifespan(app: FastAPI) -> AsyncIterator[None]:
    # Get the running event loop to pass to sync broadcast callback
    loop = asyncio.get_running_loop()
    _setup_log_broadcasting(loop)
    # Start periodic tasks and capture task references
    run_status_task, execution_history_task = await asyncio.gather(run_status(), execution_history())
    try:
        yield
    finally:
        # Clean up: remove the broadcast callback
        logger.in_memory_handler.set_broadcast_callback(None)
        # Cancel periodic tasks on shutdown
        run_status_task.cancel()
        execution_history_task.cancel()
        for task in (run_status_task, execution_history_task):
            with suppress(asyncio.CancelledError):
                await task



import sys
from collections.abc import Mapping, Sequence
from typing import (
    Any,
    ClassVar,
    cast,
)

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    model_validator,
)

from qualibrate.core.utils.exceptions import TargetsFieldNotExist
from qualibrate.core.utils.logger_m import logger
from qualibrate.core.utils.naming import get_full_class_path
from qualibrate.core.utils.parameters import recursive_properties_solver
from qualibrate.core.utils.type_protocols import TargetType
from qualibrate.core.utils.types_parsing import types_conversion

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


__all__ = [
    "ExecutionParameters",
    "GraphElementsParameters",
    "GraphParameters",
    "NodeParameters",
    "NodesParameters",
    "RunnableParameters",
]


class RunnableParameters(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        use_attribute_docstrings=True,
    )

    @classmethod
    def serialize(cls, **kwargs: Any) -> Mapping[str, Any]:
        schema = cls.model_json_schema()
        properties = schema["properties"]
        return recursive_properties_solver(properties, schema)

    @model_validator(mode="before")
    @classmethod
    def types_conversion(cls, data: Mapping[str, Any]) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], types_conversion(data, cls.serialize()))


class TargetParameter(BaseModel):
    targets_name: ClassVar[str | None] = None

    @model_validator(mode="before")
    @classmethod
    def prepare_targets(cls, data: Mapping[str, Any]) -> Mapping[str, Any]:
        if "targets" not in data:
            return data
        targets = data.get("targets")
        if cls.targets_name is None:
            raise AssertionError("Targets specified without targets name")
        if cls.targets_name in data:
            msg = (
                f"You specified `targets` and `{cls.targets_name}` (marked as "
                f"targets name) fields. `{cls.targets_name}` will be ignored."
            )
            logger.warning(msg)
        targets = types_conversion(targets, cls.model_json_schema()["properties"].get(cls.targets_name))
        return {**data, cls.targets_name: targets}

    @model_validator(mode="after")
    def targets_exists_if_specified(self) -> Self:
        if self.targets_name not in self.__class__.model_fields:
            return self
        if self.targets is not None and not isinstance(self.targets, Sequence):
            raise AssertionError(f"Targets must be an iterable of {TargetType}")
        return self

    @property
    def targets(self) -> list[TargetType] | None:
        if self.targets_name is None or self.targets_name not in self.__class__.model_fields:
            return None
        return cast(list[TargetType], getattr(self, self.targets_name))

    @targets.setter
    def targets(self, new_targets: Sequence[TargetType]) -> None:
        if self.targets_name is None:
            return
        if self.targets_name not in self.__class__.model_fields:
            raise TargetsFieldNotExist(f"Targets name ({self.targets_name}) specified but field does not exist")
        if not isinstance(new_targets, Sequence):
            raise ValueError(f"Targets must be an iterable of {TargetType}")
        setattr(self, self.targets_name, new_targets)

    @classmethod
    def serialize_targets(
        cls,
        parameters: Mapping[str, Any],
        exclude_targets: bool = False,
    ) -> Mapping[str, Any]:
        if exclude_targets:
            return {k: {**v, "is_targets": False} for k, v in parameters.items() if k != cls.targets_name}
        else:
            return {k: {**v, "is_targets": k == cls.targets_name} for k, v in parameters.items()}


class NodeParameters(RunnableParameters, TargetParameter):
    targets_name: ClassVar[str | None] = "qubits"

    @classmethod
    def serialize(cls, exclude_targets: bool = False, **kwargs: Any) -> Mapping[str, Any]:
        return cls.serialize_targets(super().serialize(), exclude_targets)


class GraphElementsParameters(RunnableParameters):
    pass


class NodesParameters(GraphElementsParameters):
    pass


class GraphParameters(RunnableParameters, TargetParameter):
    targets_name: ClassVar[str | None] = "qubits"

    @classmethod
    def serialize(cls, exclude_targets: bool = False, **kwargs: Any) -> Mapping[str, Any]:
        return cls.serialize_targets(super().serialize(), exclude_targets)


class OrchestratorParameters(RunnableParameters):
    pass


class ExecutionParameters(RunnableParameters):
    parameters: GraphParameters = Field(default_factory=GraphParameters)
    nodes: GraphElementsParameters = Field(default_factory=GraphElementsParameters)

    @classmethod
    def serialize(cls, **kwargs: Any) -> Mapping[str, Any]:
        serialized = super().serialize()
        updated_serialized = {}
        if len(serialized) > 2:
            updated_serialized = {k: v for k, v in serialized.items() if k not in ("parameters", "nodes")}
        exclude_targets = kwargs.get("exclude_targets")
        exclude_parameters_targets = exclude_targets if exclude_targets is not None else False
        exclude_nodes_targets = exclude_targets if exclude_targets is not None else True
        parameters_class = cls.model_fields["parameters"].annotation
        if parameters_class is None:
            raise RuntimeError("Graph parameters class can't be none")
        if not issubclass(parameters_class, GraphParameters):
            raise RuntimeError(f"Graph parameters class should be subclass of {get_full_class_path(GraphParameters)}")

        updated_serialized["parameters"] = parameters_class.serialize_targets(
            serialized["parameters"], exclude_targets=exclude_parameters_targets
        )
        updated_serialized["nodes"] = {
            node_name: NodeParameters.serialize_targets(params, exclude_targets=exclude_nodes_targets)
            for node_name, params in serialized["nodes"].items()
        }

        return updated_serialized

    @property
    def targets_name(self) -> str | None:
        return self.parameters.targets_name

    @property
    def targets(self) -> list[TargetType] | None:
        return self.parameters.targets

    @targets.setter
    def targets(self, new_targets: Sequence[TargetType]) -> None:
        self.parameters.targets = new_targets



from collections.abc import Callable, Mapping, Sequence
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    Generic,
    ParamSpec,
    TypeVar,
    cast,
)

import networkx as nx

from qualibrate.core.parameters import RunnableParameters
from qualibrate.core.q_runnnable import QRunnable

if TYPE_CHECKING:
    from qualibrate.core.qualibration_graph import QualibrationGraph

GraphElementTypeVar = TypeVar(
    "GraphElementTypeVar",
    bound=QRunnable[RunnableParameters, RunnableParameters],
)

P = ParamSpec("P")
R = TypeVar("R")


def ensure_finalized(
    fn: Callable[Concatenate["QualibrationGraph[Any]", P], R],
) -> Callable[Concatenate["QualibrationGraph[Any]", P], R]:
    """Decorator to block method calls beforw finalize()."""

    @wraps(fn)
    def wrapper(self: "QualibrationGraph[Any]", *args: P.args, **kwargs: P.kwargs) -> R:
        if not self._finalized:
            raise RuntimeError(f"Cannot call {fn.__name__}() because the graph isn't finalized yet.")
        return fn(self, *args, **kwargs)

    return cast(Callable[Concatenate["QualibrationGraph[Any]", P], R], wrapper)


def ensure_not_finalized(
    fn: Callable[Concatenate["QualibrationGraph[Any]", P], R],
) -> Callable[Concatenate["QualibrationGraph[Any]", P], R]:
    """Decorator to block method calls after finalize()."""

    @wraps(fn)
    def wrapper(self: "QualibrationGraph[Any]", *args: P.args, **kwargs: P.kwargs) -> R:
        if self._finalized:
            raise RuntimeError(f"Cannot call {fn.__name__}() because the graph is already finalized.")
        return fn(self, *args, **kwargs)

    return cast(Callable[Concatenate["QualibrationGraph[Any]", P], R], wrapper)


def ensure_building(
    fn: Callable[Concatenate["QualibrationGraph[Any]", P], R],
) -> Callable[Concatenate["QualibrationGraph[Any]", P], R]:
    """Decorator to allow method calls only during build phase."""

    @wraps(fn)
    def wrapper(self: "QualibrationGraph[Any]", *args: P.args, **kwargs: P.kwargs) -> R:
        if not getattr(self, "_building", False):
            raise RuntimeError(f"Cannot call {fn.__name__}() because the graph is not in build mode.")
        return fn(self, *args, **kwargs)

    return cast(Callable[Concatenate["QualibrationGraph[Any]", P], R], wrapper)


class GraphExportMixin(Generic[GraphElementTypeVar]):
    @staticmethod
    def nx_graph_export(graph: "nx.DiGraph[GraphElementTypeVar]", node_names_only: bool = False) -> Mapping[str, Any]:
        """
        Exports the graph as a networkx adjacency list.

        Args:
            graph: networkx graph to export
            node_names_only (bool): If True, only node names are included in
                the adjacency list. Defaults to False.

        Returns:
            Mapping[str, Any]: A dictionary representing the adjacency list of
            the graph.
        """
        data = dict(nx.readwrite.adjacency_data(graph))
        for key in ("multigraph", "directed", "graph"):
            data.pop(key)
        if node_names_only:
            for node, adjacency in zip(data["nodes"], data["adjacency"], strict=False):
                node["id"] = node["id"].name
                for adj in adjacency:
                    adj["id"] = adj["id"].name
        return data

    @staticmethod
    def cytoscape_representation(
        serialized: Mapping[str, Any],
    ) -> Sequence[Mapping[str, Any]]:
        """
        Returns a Cytoscape-compatible representation of the graph.

        This method generates nodes and edges in a format that can be used with
        Cytoscape for visualization purposes.

        Args:
            serialized (Mapping[str, Any]): Serialized representation of the
                graph.

        Returns:
            Sequence[Mapping[str, Any]]: List of nodes and edges for Cytoscape
                visualization.
        """
        nodes = [
            {
                "group": "nodes",
                "data": {"id": node},
                "position": {"x": 100, "y": 100},
            }
            for node in serialized["nodes"]
        ]
        edges = [
            {
                "group": "edges",
                "data": {
                    "id": f"{source}_{dest}",
                    "source": source,
                    "target": dest,
                },
            }
            for source, dest in serialized["connectivity"]
        ]
        return [*nodes, *edges]



from collections.abc import Mapping
from typing import Annotated, Any, cast

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException, status
from pydantic import BaseModel

from qualibrate.runner.api.dependencies import get_graph_nocopy as get_qgraph
from qualibrate.runner.api.dependencies import get_node_copy as get_qnode_copy
from qualibrate.runner.api.dependencies import (
    get_state,
)
from qualibrate.runner.config import (
    State,
)
from qualibrate.runner.core.run_job import (
    run_node,
    run_workflow,
    validate_input_parameters,
)
from qualibrate.runner.core.types import QGraphType, QNodeType

submit_router = APIRouter(prefix="/submit")


def _recursive_clear_node_parameters(
    parameters: Mapping[str, Any],
) -> Mapping[str, Any]:
    """
    Recursively clear run parameters of submitted elements.
    Based on qualibrate.parameters.ExecutionParameters.

    Expected out structure:
    ```python
    {
        "parameters": {
            "p1": ...
        },
        "nodes": {
            "subgraph1": {
                "parameters": {
                    "p2": ...
                },
                "nodes": {
                    "subgraph_node1": {
                        "p3": 120,
                    },
                    "subgraph_node2": {
                        "p5": 100,
                    }
                }
            },
            "outer_graph_node1": {
                "p7": 120,
            }
        }
    }
    ```
    """
    if "nodes" in parameters and "parameters" in parameters:
        return {
            "parameters": parameters["parameters"],
            "nodes": {name: _recursive_clear_node_parameters(params) for name, params in parameters["nodes"].items()},
        }
    if "parameters" in parameters:
        return dict(parameters["parameters"])
    return parameters


def clear_input_parameters(
    input_parameters: Mapping[str, Any],
) -> Mapping[str, Any]:
    return _recursive_clear_node_parameters(input_parameters)


@submit_router.post("/node")
def submit_node_run(
    input_parameters: Annotated[Mapping[str, Any], Depends(clear_input_parameters)],
    state: Annotated[State, Depends(get_state)],
    node: Annotated[QNodeType, Depends(get_qnode_copy)],
    background_tasks: BackgroundTasks,
) -> str:
    if state.is_running:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_CONTENT,
            detail="Already running",
        )
    validate_input_parameters(cast(type[BaseModel], node.parameters_class), input_parameters)
    background_tasks.add_task(run_node, node, input_parameters, state)
    return f"Node job {node.name} is submitted"


@submit_router.post("/workflow")
def submit_workflow_run(
    input_parameters: Annotated[Mapping[str, Any], Depends(clear_input_parameters)],
    state: Annotated[State, Depends(get_state)],
    graph: Annotated[QGraphType, Depends(get_qgraph)],
    background_tasks: BackgroundTasks,
) -> str:
    if state.is_running:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_CONTENT,
            detail="Already running",
        )
    validate_input_parameters(graph.full_parameters_class, input_parameters)
    background_tasks.add_task(run_workflow, graph, input_parameters, state)
    return f"Workflow job {graph.name} is submitted"



"""
13 - Full single-qubit characterization

This example demonstrates:
- A complex calibration workflow for characterizing a single qubit.
- Nested graphs to group related calibration steps.
- A combination of sequential execution, branching, and looping.
"""

from qualibrate.core import GraphParameters, QualibrationGraph, QualibrationLibrary

# Get the active library of pre-defined calibration nodes
library = QualibrationLibrary.get_active_library()


class FullCharacterizationParameters(GraphParameters):
    """Parameters for the full characterization graph."""

    qubits: list[str] = ["q1"]


# Use the context manager to build the main graph
with QualibrationGraph.build(
    "13_full_qubit_characterization",
    parameters=FullCharacterizationParameters(),
    # description="A full characterization routine for a single qubit.",
) as graph:
    # 1. Initial qubit spectroscopy to find the resonance frequency
    spectroscopy_node = library.nodes["01_demo_qubit_spectroscopy"]
    graph.add_node(spectroscopy_node)

    # 2. Gate optimization (nested subgraph)
    # This subgraph calibrates the qubit drive pulses.
    with QualibrationGraph.build(
        "gate_optimization",
        parameters=FullCharacterizationParameters(),
        # description="A subgraph to optimize gate pulses (Rabi and refined Rabi).",
    ) as gate_subgraph:
        # Start with a coarse Rabi experiment
        rabi_node = library.nodes["02_demo_rabi"]
        gate_subgraph.add_node(rabi_node)

        # Retry the Rabi node 2 times.
        gate_subgraph.loop(rabi_node, max_iterations=2)

        # Follow up with a refined Rabi for better accuracy
        refined_rabi_node = library.nodes["04_demo_rabi_refined"]
        gate_subgraph.add_node(refined_rabi_node)
        gate_subgraph.connect(rabi_node, refined_rabi_node)

    # Add the subgraph to the main graph
    graph.add_node(gate_subgraph)
    graph.connect(spectroscopy_node, gate_subgraph)

    # 3. Coherence characterization (nested subgraph)
    # This subgraph measures the qubit's coherence times (T1 and T2*).
    with QualibrationGraph.build(
        "coherence_characterization",
        parameters=FullCharacterizationParameters(),
        # description="A subgraph to measure qubit coherence (T1 and Ramsey).",
    ) as coherence_subgraph:
        t1_node = library.nodes["06_demo_t1"]
        coherence_subgraph.add_node(t1_node)

        ramsey_node = library.nodes["05_demo_ramsey"]
        coherence_subgraph.add_node(ramsey_node)

        # Run T1 and Ramsey in parallel (no connection between them)
        # The orchestrator will run nodes with no dependencies concurrently if possible.

    # Add the coherence subgraph to the main graph
    graph.add_node(coherence_subgraph)
    graph.connect(gate_subgraph, coherence_subgraph)

    # 4. Final gate fidelity assessment
    # Run Randomized Benchmarking to assess the quality of the calibrated gates.
    rb_node = library.nodes["07_demo_randomized_benchmarking"]
    rb_node.name = "final_rb_assessment"
    graph.add_node(rb_node)
    graph.connect(coherence_subgraph, rb_node)

    # 5. Error handling for the entire gate optimization step
    # If the `gate_optimization` subgraph fails completely, we can define a
    # fallback path. For this example, we just connect it to the end.
    # A real implementation might connect to a notification or reset procedure.
    graph.connect_on_failure(gate_subgraph, rb_node)


# The graph is now finalized.
if __name__ == "__main__":
    # It represents a complete, multi-stage calibration routine with nested
    # logic and failure handling.
    print(f"Running graph: {graph.name}...")
    result = graph.run()
    print("Graph execution finished.")
    print(result)



from collections.abc import Callable, Mapping
from functools import partial
from typing import Any
from urllib.parse import urljoin

import requests
from fastapi import HTTPException
from pydantic import HttpUrl
from qualibrate_config.models import (
    QualibrateConfig,
    QualibrateRunnerRemoteServiceConfig,
)

HTTPException422 = partial(HTTPException, status_code=422)


def get_runner_config(
    config: QualibrateConfig,
) -> QualibrateRunnerRemoteServiceConfig:
    if config.runner is None:
        raise RuntimeError("Qualibrate runner remote service is not configured")
    return config.runner


def request_with_db(
    path: str,
    *,
    params: Mapping[str, Any] | None = None,
    db_name: str,
    timeout: float,
    method: Callable[..., requests.Response] = requests.get,
    host: HttpUrl | str,
    **kwargs: Any,
) -> requests.Response:
    if params is None:
        params = {"db_name": db_name}
    else:
        params = dict(params)
        params["db_name"] = db_name
    result = method(
        urljoin(str(host), path),
        params=params,
        timeout=timeout,
        **kwargs,
    )
    return result



from inspect import isclass
from typing import Any


def get_full_class_path(cls_or_obj: Any) -> str:
    """Returns the full path of a class, including the module name.

    Args:
        cls_or_obj: The class to get the path of.

    Returns:
        The full path of the class or object. Generally this is of the form
        "module_name.class_name".
    """
    if not isclass(cls_or_obj):
        cls_or_obj = cls_or_obj.__class__
    class_name = cls_or_obj.__qualname__
    module_name = cls_or_obj.__module__
    if module_name in ("__main__", "builtins") or module_name is None:
        return str(class_name)
    return f"{module_name}.{class_name}"



from collections.abc import Callable, Generator, Iterator, Mapping, Sequence
from itertools import chain
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
)

from qualibrate.app.api.core.models.snapshot import (
    MachineSearchResults,
    SnapshotSearchResult,
)

if TYPE_CHECKING:
    from qualibrate.app.api.core.domain.bases.snapshot import SnapshotBase


def _get_subpath_value_wildcard(
    obj: Mapping[str, Any] | Sequence[Any],
    target_path: Sequence[str | int],
    current_path: list[str | int],
) -> Sequence[MachineSearchResults]:
    if len(target_path) == 1:
        if isinstance(obj, Sequence):
            return [MachineSearchResults(key=current_path + [i], value=val) for i, val in enumerate(obj)]
        elif isinstance(obj, Mapping):
            return [MachineSearchResults(key=current_path + [key], value=val) for key, val in obj.items()]
        else:
            return []

    if not isinstance(obj, (Mapping, Sequence)):
        return []
    iter_function = cast(
        Callable[..., Sequence[tuple[str, Any]] | Sequence[tuple[int, Any]]],
        enumerate if isinstance(obj, Sequence) else dict.items,
    )
    return list(
        chain.from_iterable(
            get_subpath_value(value, target_path[1:], current_path + [idx]) for idx, value in iter_function(obj)
        )
    )


def _check_key_valid(obj: Any, key: str | int) -> bool:
    if isinstance(obj, Sequence):
        return isinstance(key, int) and 0 <= key < len(obj)
    if isinstance(obj, Mapping):
        return isinstance(key, str) and key in obj
    return False


def get_subpath_value_mapping(
    obj: Mapping[str, Any],
    target_path: Sequence[str | int],
    current_path: list[str | int],
    key: str,
) -> Sequence[MachineSearchResults]:
    if len(target_path) == 1:
        return [MachineSearchResults(key=current_path + [key], value=obj[key])]
    else:
        return get_subpath_value(obj[key], target_path[1:], current_path + [key])


def get_subpath_value_sequence(
    obj: Sequence[Any],
    target_path: Sequence[str | int],
    current_path: list[str | int],
    key: int,
) -> Sequence[MachineSearchResults]:
    if len(target_path) == 1:
        return [MachineSearchResults(key=current_path + [key], value=obj[key])]
    else:
        return get_subpath_value(obj[key], target_path[1:], current_path + [key])


def get_subpath_value(
    obj: Mapping[str, Any] | Sequence[Any],
    target_path: Sequence[str | int],
    current_path: list[str | int] | None = None,
) -> Sequence[MachineSearchResults]:
    if current_path is None:
        current_path = []
    if len(target_path) == 0:
        return []
    key = target_path[0]
    if key == "*":
        return _get_subpath_value_wildcard(obj, target_path, current_path)
    key_valid = _check_key_valid(obj, key)
    if not key_valid:
        return []
    if isinstance(obj, Mapping):
        return get_subpath_value_mapping(obj, target_path, current_path, cast(str, key))
    return get_subpath_value_sequence(obj, target_path, current_path, cast(int, key))


def get_subpath_value_on_any_depth(
    obj: Mapping[str, Any] | Sequence[Any],
    key: str,
    current_path: list[str | int] | None = None,
    paths: list[MachineSearchResults] | None = None,
) -> list[MachineSearchResults]:
    if current_path is None:
        current_path = []
    if paths is None:
        paths = []
    if isinstance(obj, Mapping):
        for k, v in obj.items():
            if k == key:
                paths.append(MachineSearchResults(key=current_path + [k], value=v))
            else:
                get_subpath_value_on_any_depth(v, key, current_path + [k], paths)
    elif isinstance(obj, Sequence) and not isinstance(obj, str):
        for i, item in enumerate(obj):
            get_subpath_value_on_any_depth(item, key, current_path + [i], paths)
    return paths


SnapshotType = TypeVar("SnapshotType", bound="SnapshotBase")


def _get_search_results(snapshot: SnapshotType, data_path: Sequence[str | int]) -> Sequence[MachineSearchResults]:
    """Get all search results for a given data path.

    Args:
        snapshot: The snapshot to search
        data_path: The path to search for

    Returns:
        A sequence of all matching search results (empty sequence if none found)
    """
    search_results = snapshot.search(data_path, load=True)
    return search_results if search_results else []


def _get_snapshot_search_result(
    snapshot: SnapshotType,
    result: MachineSearchResults | None,
) -> SnapshotSearchResult:
    result_dict = {} if result is None else result.model_dump()
    snapshot_dump = snapshot.dump()
    return SnapshotSearchResult(snapshot=snapshot_dump, **result_dict)


def search_snapshots_data_with_filter_ascending(
    snapshots: Iterator[SnapshotType],
    data_path: Sequence[str | int],
    filter_no_change: bool,
) -> Generator[SnapshotSearchResult, None, None]:
    previous_search_result: MachineSearchResults | None = None
    for snapshot in snapshots:
        search_results = _get_search_results(snapshot, data_path)
        for search_result in search_results:
            if filter_no_change:
                if previous_search_result != search_result:
                    previous_search_result = search_result
                    yield _get_snapshot_search_result(snapshot, previous_search_result)
            else:
                yield _get_snapshot_search_result(snapshot, search_result)


def search_snapshots_data_with_filter_descending(
    snapshots: Iterator[SnapshotType],
    data_path: Sequence[str | int],
    filter_no_change: bool,
) -> Generator[SnapshotSearchResult, None, None]:
    snapshot = next(snapshots, None)
    if snapshot is None:
        return
    search_results = _get_search_results(snapshot, data_path)
    if not filter_no_change:
        for search_result in search_results:
            yield _get_snapshot_search_result(snapshot, search_result)
        for snapshot in snapshots:
            search_results = _get_search_results(snapshot, data_path)
            for search_result in search_results:
                yield _get_snapshot_search_result(snapshot, search_result)
        return
    previous_snapshot: SnapshotType = snapshot
    previous_results: Sequence[MachineSearchResults] = search_results
    for snapshot in snapshots:
        search_results = _get_search_results(snapshot, data_path)
        if previous_results != search_results:
            for search_result in previous_results:
                yield _get_snapshot_search_result(previous_snapshot, search_result)
        previous_snapshot = snapshot
        previous_results = search_results
    for search_result in previous_results:
        yield _get_snapshot_search_result(previous_snapshot, search_result)



from typing import Annotated

from pydantic import BaseModel, Field

from qualibrate.runner.core.models.common import RunError
from qualibrate.runner.core.models.enums import RunStatusEnum


class WorkflowStatus(BaseModel):
    """Model representing the status of a workflow execution."""

    active: Annotated[
        bool,
        Field(description="Indicates if the workflow is currently running."),
    ]
    status: Annotated[RunStatusEnum, Field(description="The current status of the workflow.")]
    active_node_name: Annotated[
        str | None,
        Field(description="The name of the currently active node, if any."),
    ]
    nodes_completed: Annotated[
        int,
        Field(
            description=("The number of nodes that have been completed in the workflow."),
        ),
    ]
    nodes_total: Annotated[int, Field(description="The total number of nodes in the workflow.")]
    run_duration: Annotated[
        float,
        Field(
            description=("The total runtime duration of the workflow in seconds."),
        ),
    ]
    error: Annotated[
        RunError | None,
        Field(
            description=("Any error encountered during the workflow execution, if applicable."),
        ),
    ] = None
